import traceback
from functools import partial
from typing import List, Optional, Tuple, Union, overload  # noqa

import lndb
import sqlmodel as sqm
from lamin_logger import logger
from lndb import settings as setup_settings
from lndb_storage import delete_storage, store_object, write_adata_zarr
from lndb_storage._file import print_hook
from lnschema_core import File
from lnschema_core.dev._storage import storage_key_from_file
from pydantic.fields import ModelPrivateAttr

from .._docs import doc_args
from ._core import file_to_sqm, get_session_from_kwargs
from ._select import select

add_docs = """
Insert or update data records.

Inserts a new :term:`record` if the corresponding row doesn't exist.
Updates the corresponding row with the record if it exists.

To update a row, query it with `.select` and modify it before
passing it to `add`.

Guide: :doc:`/guide/add-delete`.

Args:
    record: One or multiple records as instances of `SQLModel`.

Returns:
    The record as returned from the database with a `created_at` timestamp.

Examples:

    Add a record (errors if already exists):

    >>> ln.add(ln.Transform(name="My pipeline"))
    Transform(id="0Cb86EZj", name="My pipeline", ...)

    Update an existing record:

    >>> transform = ln.select(ln.Transform, id="0Cb86EZj").one()
    >>> transform.name = "New name"
    >>> ln.add(transform)
    Transform(id="0Cb86EZj", name="New name", ...)

    Add a record with passed fields if not yet exists:

    >>> # add a record if the metadata combination is not already exist in the DB
    >>> # if exists, returns the existing record from the DB
    >>> ln.add(ln.Transform, name="My transform", v="1")
    Transform(id="0Cb86EZj", name="My pipeline", ...)
    >>> # is equivalent to the following:
    >>> transform = ln.select(ln.Transform, name="My transform", v="1").one_or_none()
    >>> if transform is None:
    >>>     ln.add(transform)

"""


@overload
def add(record: sqm.SQLModel) -> sqm.SQLModel:
    ...


# Currently seeing the following error without type ignore:
# Overloaded function signature 2 will never be matched: signature 1's parameter
# type(s) are the same or broader
@overload
def add(records: List[sqm.SQLModel]) -> List[sqm.SQLModel]:  # type: ignore
    ...


@overload
def add(  # type: ignore
    entity: sqm.SQLModel, **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    ...


@doc_args(add_docs)
def add(  # type: ignore
    record: Union[sqm.SQLModel, List[sqm.SQLModel]], **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    """{}"""  # noqa
    session = get_session_from_kwargs(fields)
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = file_to_sqm(record)
        results = select(model, **fields).one_or_none()
        if results is None:
            records = [model(**fields)]
        else:
            logger.info(
                f"An existing {results.__class__.__name__} record is found in the DB:"
            )
            return results

    if session is None:  # assume global session
        session = setup_settings.instance.session()
        setup_settings.instance._cloud_sqlite_locker.lock()
        close = True
    else:
        close = False

    # commit metadata to database
    db_error = None
    for record in records:
        # the following ensures that queried objects (within __init__)
        # behave like queried objects, only example right now: Run
        if (
            _private_not_empty(record, "_ln_identity_key")
            and record._ln_identity_key is not None  # noqa
        ):
            record._sa_instance_state.key = record._ln_identity_key
        session.add(record)
    try:
        session.commit()
    except Exception as e:
        db_error = e

    # upload data objects to storage
    added_records = []
    if db_error is None:
        added_records, upload_error = upload_committed_records(records, session)

    if close:
        session.close()
        setup_settings.instance._update_cloud_sqlite_file()
        setup_settings.instance._cloud_sqlite_locker.unlock()

    error = db_error or upload_error
    if error is not None:
        error_message = prepare_error_message(records, added_records, error)
        raise RuntimeError(error_message)
    elif len(added_records) > 1:
        return added_records
    else:
        return added_records[0]


def upload_committed_records(records, session):
    """Upload records in a list of database-commited records to storage.

    If any upload fails, subsequent records are cleaned up from the DB.
    """
    # make sure ALL records are up-to-date to enable accurate comparison
    # during metadata cleanup
    error = None
    added_records = []
    for record in records:
        session.refresh(record)

    # upload data objects
    for record in records:
        if isinstance(record, File) and _private_not_empty(record, "_local_filepath"):
            try:
                upload_data_object(record)
            except Exception as e:
                error = e
                break
        added_records += [record]

    # clear old files on update
    for record in added_records:
        if isinstance(record, File) and _private_not_empty(record, "_clear_storagekey"):
            try:
                if record._clear_storagekey is not None:
                    delete_storage(record._clear_storagekey)
                    record._clear_storagekey = None
            except Exception as e:
                error = e

    # clean up metadata for objects not uploaded to storage
    if error is not None:
        for record in records:
            if record not in added_records:
                session.delete(record)
        session.commit()

    return added_records, error


def prepare_error_message(records, added_records, error) -> str:
    if len(records) == 1 or len(added_records) == 0:
        error_message = (
            "No entries were uploaded or committed"
            " to the database. See error message:\n\n"
        )
    else:
        error_message = (
            "The following entries have been"
            " successfully uploaded and committed to the database:\n"
        )
        for record in added_records:
            error_message += (
                f"- {', '.join(record.__repr__().split(', ')[:3]) + ', ...)'}\n"
            )
        error_message += "\nSee error message:\n\n"
    error_message += f"{str(error)}\n\n{traceback.format_exc()}"
    return error_message


def upload_data_object(file) -> None:
    """Store and add file and its linked entries."""
    # do NOT hand-craft the storage key!
    file_storage_key = storage_key_from_file(file)
    if (
        _private_not_empty(file, "_to_store")
        and file._to_store
        and file.suffix != ".zarr"
    ):
        logger.hint(f"storing object {file.name} with key {file_storage_key}")
        store_object(file._local_filepath, file_storage_key)
    elif (
        file.suffix == ".zarr"
        and _private_not_empty(file, "_memory_rep")
        and file._memory_rep is not None
    ):
        logger.hint(f"storing object {file.name} with key {file_storage_key}")
        storagepath = lndb.settings.storage.key_to_filepath(file_storage_key)
        print_progress = partial(print_hook, filepath=file.name)
        write_adata_zarr(file._memory_rep, storagepath, callback=print_progress)


# sometimes obj.attr results in
def _private_not_empty(obj, attr):
    if hasattr(obj, attr):
        return not isinstance(getattr(obj, attr), ModelPrivateAttr)
    else:
        return False



"""v0.14.0.

Revision ID: cf5913791674
Revises: 2ddcb037e3ea
Create Date: 2022-10-19 19:21:27.015737

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cf5913791674"
down_revision = "2ddcb037e3ea"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "dset",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["created_by"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("dset", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_dset_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_dset_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_dset_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_dset_updated_at"), ["updated_at"], unique=False)

    op.create_table(
        "project",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["created_by"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_project_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_project_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_project_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_project_updated_at"), ["updated_at"], unique=False)

    op.create_table(
        "project_dset",
        sa.Column("project_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dset_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dset_id"],
            ["dset.id"],
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.PrimaryKeyConstraint("project_id", "dset_id"),
    )

    op.create_table(
        "dset_dobject",
        sa.Column("dset_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dobject_id"],
            ["dobject.id"],
        ),
        sa.ForeignKeyConstraint(
            ["dset_id"],
            ["dset.id"],
        ),
        sa.PrimaryKeyConstraint("dset_id", "dobject_id"),
    )


def downgrade() -> None:
    pass



from typing import Optional

import pandas as pd
import sqlmodel as sqm
from lndb import settings
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlmodel.main import SQLModelMetaclass

from .._docs import doc_args
from ._core import file_to_sqm, get_session_from_kwargs

select_docs = """
Query data.

Guide: :doc:`/guide/select`.

Args:
    entity: Table, tables, or tables including column specification.
    fields: Fields and values passed as keyword arguments.

Returns:
    A :class:`~lamindb.dev.db.SelectStmt` object.
"""


@doc_args(select_docs)
def select(*entity: sqm.SQLModel, **fields) -> "SelectStmt":
    """{}"""
    session = get_session_from_kwargs(fields)
    # if ln.File is passed, replace it with File SQLModel class
    entities = file_to_sqm(entity)

    # continue with user-facing variables
    if len(entities) > 1 and len(fields) > 0:
        raise RuntimeError("Can only pass fields for a single entity.")
    elif len(fields) > 0:
        # was in `get` before, but there it leads to an inhomogeneous return type
        conditions = [getattr(entities[0], k) == v for k, v in fields.items()]
        return SelectStmt(*entities, session=session).where(*conditions)
    return SelectStmt(*entities, session=session)


def to_df(*entities, result):
    if len(result) == 0:
        if len(entities) == 1:
            entity = entities[0]
            return pd.DataFrame(columns=entity.__fields__)
        else:
            # not implemented
            return pd.DataFrame()
    row = result[0]
    if len(entities) == 1:
        entity = entities[0]
        if isinstance(entity, SQLModelMetaclass):
            records = [row.dict() for row in result]
            df = pd.DataFrame.from_records(
                records, columns=entity.__fields__, coerce_float=True
            )
            if "id" in df.columns:
                if "version" in df.columns:
                    df = df.set_index(["id", "version"])
                else:
                    df = df.set_index("id")
            return df
        elif hasattr(entity, "class_") and isinstance(entity.class_, SQLModelMetaclass):
            return pd.DataFrame(result)
        else:
            raise RuntimeError(
                "Unknown entities. Pass either a SQLModel or a field of a SQLModel."
            )
    if len(entities) > 1:
        records = []
        keys_types = []
        for entity in entities:
            if isinstance(entity, SQLModelMetaclass):
                keys_types.append((entity.__table__.name, 0))
            elif hasattr(entity, "class_") and isinstance(
                entity.class_, SQLModelMetaclass
            ):
                keys_types.append((entity.class_.__table__.name, 1))
            else:
                raise RuntimeError(
                    "Unknown entities. Pass either a SQLModel or a field of a SQLModel."
                )
        for row in result:
            record = {}
            for i, (key, type) in enumerate(keys_types):
                if type == 0:
                    sub_record = {
                        (key, field): getattr(row[i], field)
                        for field in row[i].__fields__
                    }
                    record.update(sub_record)
                else:
                    k = entities[i].expression.name
                    record[(key, k)] = row[i]
            records.append(record)
        df = pd.DataFrame(records)
        df.columns = pd.MultiIndex.from_tuples(df.columns)
        return df


class ExecStmt:
    """Executable statement."""

    def __init__(self, *, tables, stmt, session: Optional[sqm.Session] = None):
        self._stmt = stmt
        self._tables = tables
        self._result = None
        self._session: Optional[sqm.Session] = session

    def _execute(self):
        # cache the query result for the lifetime of the object
        if self._result is None:
            if self._session is None:  # use global session
                ss = settings.instance.session()
                close = True
            else:
                ss = self._session
                close = False
            with ss.no_autoflush:
                self._result = ss.exec(self._stmt).all()
            if close:
                ss.close()

    def all(self):
        """Return all result as a list."""
        self._execute()
        return self._result

    def df(self):
        """Return result as a DataFrame.

        The DataFrame will have MultiIndex columns if multiple entities are passed.
        """
        self._execute()
        return to_df(*self._tables, result=self._result)

    def first(self):
        """Return the first result of select or None if no result are found."""
        self._execute()
        if len(self._result) == 0:
            return None
        return self._result[0]

    def one(self):
        """Return exactly one result or raise an exception."""
        self._execute()
        if len(self._result) == 0:
            raise NoResultFound
        elif len(self._result) > 1:
            raise MultipleResultsFound
        else:
            return self._result[0]

    def one_or_none(self):
        """Return at most one result or raise an exception."""
        self._execute()
        if len(self._result) == 0:
            return None
        elif len(self._result) == 1:
            return self._result[0]
        else:
            raise MultipleResultsFound


class SelectStmt(ExecStmt):
    """Executable select statement.

    Offers an exact subset of SQLAlchemy of constraining select with `where`,
    `join`, `order_by`, `offset`, `limit`.

    Offers accessing the results of the executed statement via `all`, `df`,
    `one`, `one_or_none` through the base class `ExecStmt`.
    """

    def __init__(self, *tables, stmt=None, session=None) -> None:
        self._tables = tables
        self._session = session
        if stmt is None:
            stmt = sqm.select(*tables)
        super().__init__(tables=tables, stmt=stmt, session=session)

    def where(self, *conditions):
        """Pass one or multiple conditions.

        If multiple conditions are passed, they are combined using AND.

        If OR is desired, use `sqlmodel.or_`.
        """
        return SelectStmt(
            *self._tables, stmt=self._stmt.where(*conditions), session=self._session
        )

    def join(self, *expression, **fields):
        """Pass a target table as an expression."""
        stmt = self._stmt.join(*expression)
        if len(fields) > 0 and len(expression) == 1:
            conditions = [getattr(expression[0], k) == v for k, v in fields.items()]
            stmt = stmt.where(*conditions)
        elif len(fields) > 0 and len(expression) > 1:
            raise RuntimeError("Can only pass fields for a single entity.")
        return SelectStmt(*self._tables, stmt=stmt, session=self._session)

    def order_by(self, expression):
        """Pass a field."""
        return SelectStmt(
            *self._tables, stmt=self._stmt.order_by(expression), session=self._session
        )

    def offset(self, n):
        """Pass an integer."""
        return SelectStmt(
            *self._tables, stmt=self._stmt.offset(n), session=self._session
        )

    def limit(self, n):
        """Pass an integer."""
        return SelectStmt(
            *self._tables, stmt=self._stmt.limit(n), session=self._session
        )



from pathlib import Path

import nbproject_test as test
from lamin_logger import logger


def test_notebooks():
    # assuming this is in the tests folder
    docsdir = Path(__file__).parents[1] / "docs/"

    for subdir in ["guide", "faq"]:
        checkdir = docsdir / subdir
        logger.info(f"\n---{checkdir.stem}---")
        test.execute_notebooks(checkdir, write=True)



"""v0.25.6.

Revision ID: 9d283a1685a5
Revises: f6b6b85cdffc
Create Date: 2023-02-02 21:22:35.229193

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9d283a1685a5"
down_revision = "f6b6b85cdffc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    if sqlite:
        with op.batch_alter_table(f"{prefix}dobject", schema=schema) as batch_op:
            batch_op.add_column(sa.Column("_filekey", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
            batch_op.create_index(batch_op.f("ix_core.dobject__filekey"), ["_filekey"], unique=False)
    else:
        op.add_column(
            f"{prefix}dobject",
            sa.Column("_filekey", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
            schema=schema,
        )
        op.create_index(
            op.f("ix_core_dobject__filekey"),
            "dobject",
            ["_filekey"],
            unique=False,
            schema=schema,
        )


def downgrade() -> None:
    pass



import operator
from typing import Union

# todo: add all operators
BINARY_OPS = [
    "__add__",
    "__mul__",
    "__or__",
    "__and__",
    "__eq__",
    "__ge__",
    "__gt__",
    "__le__",
    "__lt__",
    "__matmul__",
    "__pow__",
    "__truediv__",
    "__floordiv__",
]
UNARY_OPS = ["__abs__", "__neg__", "__invert__"]


class MetaCatchOperators(type):
    """Catch operators and return `LazyOperator`."""

    def __new__(meta, name, bases, dct):
        cls_obj = super().__new__(meta, name, bases, dct)

        for op in BINARY_OPS:
            op_func = getattr(operator, op)

            def left(self, right, op_func=op_func):
                return LazyOperator(self, right, op_func)

            setattr(cls_obj, op, left)

            def right(self, left, op_func=op_func):
                return LazyOperator(left, self, op_func)

            setattr(cls_obj, f"__r{op[2:]}", right)

        for op in UNARY_OPS:
            op_func = getattr(operator, op)

            def unary(self, op_func=op_func):
                return LazyOperator(self, None, op_func)

            setattr(cls_obj, op, unary)

        return cls_obj


class CatchAccess:
    def __getattr__(self, prop):
        """Catch a property and return `LazyProperty`."""
        return LazyProperty(self, prop)

    def __array_function__(self, func, types, args, kwargs):
        """Catch a numpy function and return `LazyNumpyFunc`."""
        return LazyNumpyFunc(func, args, kwargs)


class LazyOperator(CatchAccess, metaclass=MetaCatchOperators):
    def __init__(self, left, right, op):
        self._left = left
        self._right = right
        self._op = op

    def evaluate(self, **kwargs):
        _left_eval = self._left
        if hasattr(_left_eval, "evaluate"):
            _left_eval = _left_eval.evaluate(**kwargs)

        _right_eval = self._right
        if hasattr(_right_eval, "evaluate"):
            _right_eval = _right_eval.evaluate(**kwargs)

        if _right_eval is None:
            return self._op(_left_eval)
        else:
            return self._op(_left_eval, _right_eval)


class LazyProperty(CatchAccess, metaclass=MetaCatchOperators):
    def __init__(self, obj, name):
        self._name = name
        self._obj = obj
        self._call = False

        self._args = tuple()
        self._kwargs = dict()

    def evaluate(self, **kwargs):
        obj = self._obj
        if hasattr(self._obj, "evaluate"):
            obj = self._obj.evaluate(**kwargs)

        attr = getattr(obj, self._name)
        if self._call:
            result = attr(*self._args, **self._kwargs)
        else:
            result = attr
        return result

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._call = True

        return self


class LazyNumpyFunc(CatchAccess, metaclass=MetaCatchOperators):
    def __init__(self, func, args, kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def evaluate(self, **kwargs):
        args_eval = []
        for arg in self._args:
            if hasattr(arg, "evaluate"):
                arg = arg.evaluate(**kwargs)
            args_eval.append(arg)

        kwargs_eval = {}
        for key, val in self._kwargs.items():
            if hasattr(val, "evaluate"):
                val = val.evaluate(**kwargs)
            kwargs_eval[key] = val

        return self._func(*args_eval, **kwargs_eval)


class LazyField(CatchAccess, metaclass=MetaCatchOperators):
    def __init__(self, name, as_attr=True):
        self.name = name
        self._as_attr = as_attr

    def evaluate(self, obj):
        if self._as_attr:
            return getattr(obj, self.name)
        else:
            return obj[self.name]


class Lazy:
    def __getattr__(self, attr):
        """Get `LazyField` as an attribute."""
        return LazyField(attr, as_attr=True)

    def __getitem__(self, key):
        """Get `LazyField` as an item."""
        return LazyField(key, as_attr=False)


lazy = Lazy()
LazySelector = Union[LazyOperator, LazyProperty, LazyNumpyFunc]



from itertools import islice
from pathlib import Path
from typing import Optional, Union

import lndb
from lamin_logger import logger
from lndb_storage import UPath
from lnschema_core import File, Run
from lnschema_core.dev import id as id_generator

from ._file import (
    get_check_path_in_storage,
    get_relative_path_to_directory,
    get_relative_path_to_root,
)


def get_folder_kwargs_from_data(
    path: Union[Path, UPath, str],
    *,
    name: Optional[str] = None,
    key: Optional[str] = None,
    run: Optional[Run] = None,
):
    folderpath = UPath(path)
    check_path_in_storage = get_check_path_in_storage(folderpath)

    if key is None and check_path_in_storage:
        folder_key = get_relative_path_to_root(path=folderpath).as_posix()
    elif key is None:
        folder_key = id_generator.folder()
    else:
        folder_key = key

    # always sanitize by stripping a trailing slash
    folder_key = folder_key.rstrip("/")

    logger.hint(f"using storage key = {folder_key}")

    # TODO: UPath doesn't list the first level files and dirs with "*"
    pattern = "" if isinstance(folderpath, UPath) else "*"

    files = []
    for filepath in folderpath.rglob(pattern):
        if filepath.is_file():
            relative_path = get_relative_path_to_directory(filepath, folderpath)
            file_key = folder_key + "/" + relative_path.as_posix()
            files.append(File(filepath, run=run, key=file_key))

    logger.hint(f"-> n_files = {len(files)}")

    kwargs = dict(
        name=folderpath.name if name is None else name,
        key=folder_key,
        storage_id=lndb.settings.storage.id,
        files=files,
    )
    privates = dict(
        local_filepath=folderpath if isinstance(folderpath, UPath) else None,
        cloud_filepath=None if isinstance(folderpath, UPath) else folderpath,
    )
    return kwargs, privates


# exposed to users as Folder.tree()
def tree(
    dir_path: Union[Path, UPath, str],
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
):
    """Given a directory Path object print a visual tree structure.

    Adapted from: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python  # noqa
    """
    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "

    dir_path = UPath(dir_path)
    files = 0
    directories = 0

    def inner(dir_path: Path, prefix: str = "", level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        # this is needed so that the passed folder is not listed
        contents = [
            i
            for i in dir_path.iterdir()
            if i.as_posix().rstrip("/") != dir_path.as_posix().rstrip("/")
        ]
        if limit_to_directories:
            contents = [d for d in contents if d.is_dir()]
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                yield prefix + pointer + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1

    print(dir_path.name)
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        print(line)
    if next(iterator, None):
        print(f"... length_limit, {length_limit}, reached, counted:")
    print(f"\n{directories} directories" + (f", {files} files" if files else ""))



"""Base classes for a lab schema in biology (`vvhc`).

Import the package::

   import lnbase_biolab

Base classes:

.. autosummary::
   :toctree: .

   ExperimentBase
   ExperimentTypeBase
   BiosampleBase
   TechsampleBase

Development tools:

.. autosummary::
   :toctree: .

   dev
   link

"""
_codeproject_id = "vvhc"
_name = "biolab"
__version__ = "0.15.0"

from . import dev
from ._core import (  # noqa
    BiosampleBase,
    ExperimentBase,
    ExperimentTypeBase,
    TechsampleBase,
)



"""v0.3.3.

Revision ID: 1c531ea346cf
Revises: 8c78543d1c5b
Create Date: 2022-08-19 15:51:31.185793

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "1c531ea346cf"
down_revision = "8c78543d1c5b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Completely new table
    op.create_table(
        "storage",
        sa.Column("root", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("region", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.Column("time_updated", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("root"),
    )

    # Add storage_root column
    op.create_table(
        "dobject_tmp",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("file_suffix", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dsource_id", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("storage_root", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.Column("time_updated", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dsource_id"],
            ["dtransform.id"],
        ),
        sa.ForeignKeyConstraint(
            ["storage_root"],
            ["storage.root"],
        ),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.execute(
        "insert into dobject_tmp(id, v, name, file_suffix, dsource_id,"
        " time_created, time_updated) select id, v, name, file_suffix, dsource_id,"
        " time_created, time_updated from dobject"
    )
    op.drop_table("dobject")
    op.rename_table("dobject_tmp", "dobject")


def downgrade() -> None:
    pass



from typing import Dict, List, Tuple, Union, overload  # noqa

import sqlmodel as sqm
from lndb import settings

from .._docs import doc_args
from ._add import add, add_docs
from ._select import SelectStmt, select, select_docs


class Session:
    """Database session.

    FAQ: :doc:`/faq/session`

    It offers `.select`, `.add` and `.delete` attached to an open session, which
    is needed for lazy loading of relationships.

    The session object should be closed when it's not longer needed and
    typically used within a `with` statement::

        with Session() as ss:
            file = ss.select(ln.File, name="My test").one()
            ...
    """

    def __init__(self):
        settings.instance._cloud_sqlite_locker.lock()
        self._session = settings.instance.session()

        self._update = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._session.__exit__(exc_type, exc_value, exc_tb)
        self._finish()

    def _finish(self):
        if self._update:
            settings.instance._update_cloud_sqlite_file()
        settings.instance._cloud_sqlite_locker.unlock()

    @doc_args(add_docs)
    def add(
        self, record: Union[sqm.SQLModel, List[sqm.SQLModel]], **fields
    ) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
        """{}"""  # noqa
        fields["session"] = self._session
        self._update = True
        return add(record=record, **fields)

    @doc_args(select_docs)
    def select(self, *entity: sqm.SQLModel, **fields) -> SelectStmt:
        """{}"""
        fields["session"] = self._session
        return select(*entity, **fields)

    def close(self):
        """Close the session."""
        self._session.close()
        self._finish()



from typing import List, Optional, Union

import nbproject as _nb
from lamin_logger import logger
from lnschema_core import Run, Transform

from ._context import context


# this whole class is deprecated, see lamindb.context instead!
class nb:
    """Manage Jupyter notebooks.

    For more background, see `nbproject <https://lamin.ai/docs/nbproject>`__.
    """

    transform: Transform = None  # Notebook of this Python session
    run: Run = None  # run of this Python session

    @classmethod
    def header(
        cls,
        *,
        pypackage: Union[str, List[str], None] = None,
        filepath: Optional[str] = None,
        env: Optional[str] = None,
        id: Optional[str] = None,
        v: Optional[str] = "0",
        name: Optional[str] = None,
    ) -> Run:
        """Track the notebook & display metadata.

        Call without arguments in most settings.

        If the notebook has no nbproject metadata, initializes & writes metadata
        to disk.

        Args:
            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            env: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
            id: Pass a notebook id manually.
            v: Pass a notebook version manually.
            name: Pass a notebook name manually.
        """
        logger.warning(
            "DeprecationWarning: Please replace ln.nb.header() with ln.track()."
        )
        context._track(pypackage=pypackage, notebook_path=filepath, editor=env)
        cls.transform = context.transform
        cls.run = context.run
        return cls.run

    @classmethod
    def publish(cls, version: str = None, i_confirm_i_saved: bool = False):
        """Publish the notebook.

        Args:
            version: Notebook version to publish.
            i_confirm_i_saved: Only relevant outside Jupyter Lab & Notebook as a
                safeguard against losing the editor buffer content.
        """
        import lamindb as ln

        if cls.transform is None:
            cls.transform = ln.context.transform
        if cls.run is None:
            cls.run = ln.context.run
        from nbproject._publish import finalize_publish, run_checks_for_publish

        result = run_checks_for_publish(
            calling_statement="publish(", i_confirm_i_saved=i_confirm_i_saved
        )
        if result != "checks-passed":
            return result
        finalize_publish(calling_statement="publish(", version=version)
        # update DB
        transform = ln.select(
            Transform, id=cls.transform.id, version=cls.transform.version  # type: ignore  # noqa
        ).one()
        # update version
        transform.title = _nb.meta.live.title
        if version != cls.transform.version:  # type: ignore
            transform_add = Transform(
                id=transform.id,
                version=version,
                name=transform.name,
                title=transform.title,
            )
        else:
            transform_add = transform
        ln.add(transform_add)
        if version != cls.transform.version:  # type: ignore
            cls.run.transform_version = version  # type: ignore
            ln.add(cls.run)
            ln.delete(transform)



"""v0.32.0."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic

revision = "6de59093e378"
down_revision = "5846a15d9241"


def upgrade() -> None:
    sqlite, prefix, schema, delim = get_sqlite_prefix_schema_delim_from_alembic()

    op.alter_column(f"{prefix}file", column_name="_objectkey", new_column_name="key", schema=schema)
    op.alter_column(f"{prefix}folder", column_name="_objectkey", new_column_name="key", schema=schema)
    with op.batch_alter_table(f"{prefix}file", schema=schema) as batch_op:
        batch_op.drop_index(f"ix_core{delim}file__objectkey")
        batch_op.drop_constraint("uq_storage__objectkey_suffix", type_="unique")
        batch_op.create_index(batch_op.f(f"ix_core{delim}file_key"), ["key"], unique=False)
        batch_op.create_unique_constraint("uq_file_storage_key", ["storage_id", "key"])

    with op.batch_alter_table(f"{prefix}folder", schema=schema) as batch_op:
        batch_op.add_column(sa.Column("storage_id", sqm.sql.sqltypes.AutoString(), nullable=True))
        batch_op.create_index(op.f(f"ix_core{delim}folder_storage_id"), ["storage_id"], unique=False)
        batch_op.drop_index(f"ix_core{delim}folder__objectkey")
        batch_op.create_index(batch_op.f(f"ix_core{delim}folder_key"), ["key"], unique=False)
        batch_op.create_unique_constraint("uq_folder_storage_key", ["storage_id", "key"])
        batch_op.create_foreign_key(op.f("fk_folder_storage_storage_id"), f"{prefix}storage", ["storage_id"], ["id"], referent_schema=schema)


def downgrade() -> None:
    pass



"""Storage → object.

File → memory:

.. autosummary::
   :toctree: .

   h5ad_to_anndata

Store files:

.. autosummary::
   :toctree: .

   store_object
   store_png
   delete_storage
"""

__version__ = "0.2rc6"  # denote a release candidate for 0.1.0 with 0.1rc1

from lndb.dev.upath import UPath
from lndb.dev.upath import infer_filesystem as _infer_filesystem

from ._file import delete_storage, load_to_memory, store_object
from ._h5ad import h5ad_to_anndata
from ._images import store_png
from ._subset import subset
from ._zarr import read_adata_zarr, write_adata_zarr



from pathlib import Path
from typing import Optional

from anndata import AnnData
from anndata import __version__ as anndata_v
from lndb import settings as setup_settings
from lndb_storage import load_to_memory
from lndb_storage.object import _subset_anndata_file
from lndb_storage.object._subset_anndata import CloudAnnData
from lnschema_core import File
from lnschema_core.dev._storage import filepath_from_file
from lnschema_core.link import RunIn
from lnschema_core.types import DataLike
from packaging import version
from sqlalchemy.orm.session import object_session

from lamindb._context import context
from lamindb.dev import LazyDataFrame

from ._settings import settings
from .dev.db._add import add as ln_add

File.__doc__ = """Files: serialized data objects.

- Guide: :doc:`/guide/track`
- FAQ: :doc:`/faq/ingest`

Args:
   data: `Union[PathLike, DataLike] = None` - A file path or an in-memory data
      object to serialize. Can be a cloud path.
   key: `Optional[str] = None` - A storage key, a relative filepath within the
      storage location, e.g., an S3 or GCP bucket.
   name: `Optional[str] = None` - A name. Defaults to a file name for a file.
   run: `Optional[Run] = None` - The generating run.
   features: `List[Features] = None` - A feature set record.
   id: `Optional[str] = None` - The id of the file. Auto-generated if not passed.
   input_of: `List[Run] = None` - Runs for which the file
      is as an input.

Often, files represent atomic datasets in object storage:
jointly measured observations of features (:class:`~lamindb.Features`).
They are generated by running code (:class:`~lamindb.Transform`),
instances of :class:`~lamindb.Run`.

Data objects often have canonical on-disk and in-memory representations. LaminDB
makes some configurable default choices (e.g., serialize a `DataFrame` as a
`.parquet` file).

Some datasets do not have a canonical in-memory representation, for
instance, `.fastq`, `.vcf`, or files describing QC of datasets.

.. note:: Examples for storage ⟷ memory correspondence:

   - Table: `.csv`, `.tsv`, `.parquet`, `.ipc`, `.feather`
     ⟷ `pd.DataFrame`, `polars.DataFrame`
   - Annotated matrix: `.h5ad`, `.h5mu`, `.zarr` ⟷ `anndata.AnnData`, `mudata.MuData`
   - Image: `.jpg`, `.png` ⟷ `np.ndarray`, ...
   - Tensor: zarr directory, TileDB store ⟷ zarr loader, TileDB loader
   - Fastq: `.fastq` ⟷ /
   - VCF: `.vcf` ⟷ /
   - QC: `.html` ⟷ /

"""


def backed(self: File) -> CloudAnnData:
    """Return a cloud-backed AnnData object for streaming."""
    if self.suffix not in (".h5ad", ".zarr"):
        raise ValueError("File should have an AnnData object as the underlying data")
    return CloudAnnData(self)


def subsetter(self: File) -> LazyDataFrame:
    """A subsetter to pass to ``.stream()``.

    Currently, this returns an instance of an
    unconstrained :class:`~lamindb.dev.LazyDataFrame`
    to be evaluated in ``.stream()``.

    In the future, this will be constrained by metadata of the file, it's
    feature- and sample-level descriptors, like `.obs`, `.var`, `.columns`, `.rows`.
    """
    return LazyDataFrame()


def stream(
    self: File,
    subset_obs: Optional[LazyDataFrame] = None,
    subset_var: Optional[LazyDataFrame] = None,
    is_run_input: Optional[bool] = None,
) -> AnnData:
    """Stream the file into memory. Allows subsetting an AnnData object.

    Args:
        subset_obs: ``Optional[LazyDataFrame] = None`` - A DataFrame query to
            evaluate on ``.obs`` of an underlying ``AnnData`` object.
        subset_var: ``Optional[LazyDataFrame] = None`` - A DataFrame query to
            evaluate on ``.var`` of an underlying ``AnnData`` object.

    Returns:
        The streamed AnnData object.

    Example:

    >>> file = ln.select(ln.File).where(...).one()
    >>> obs = file.subsetter()
    >>> obs = (
    >>>     obs.cell_type.isin(["dendritic cell", "T cell")
    >>>     & obs.disease.isin(["Alzheimer's"])
    >>> )
    >>> file.stream(subset_obs=obs, is_run_input=True)

    """
    if self.suffix not in (".h5ad", ".zarr"):
        raise ValueError("File should have an AnnData object as the underlying data")
    _track_run_input(self, is_run_input)

    if subset_obs is None and subset_var is None:
        return load_to_memory(filepath_from_file(self), stream=True)

    if self.suffix == ".h5ad" and subset_obs is not None and subset_var is not None:
        raise ValueError(
            "Can not subset along both subset_obs and subset_var at the same time"
            " for an AnnData object stored as a h5ad file."
            " Please resave your AnnData as zarr to be able to do this"
        )

    if self.suffix == ".zarr" and version.parse(anndata_v) < version.parse("0.9.1"):
        raise ValueError(
            f"anndata=={anndata_v} does not support `.subset` of zarr stored AnnData."
            " Please install anndata>=0.9.1"
        )

    return _subset_anndata_file(self, subset_obs, subset_var)


def _track_run_input(file: File, is_run_input: Optional[bool] = None):
    if is_run_input is None:
        track_run_input = settings.track_run_inputs_upon_load
    else:
        track_run_input = is_run_input
    if track_run_input:
        if context.run is None:
            raise ValueError(
                "No global run context set. Call ln.context.track() or link input to a"
                " run object via `run.inputs.append(file)`"
            )
        if object_session(file) is None:
            # slower, no session open, doesn't use relationship
            run_in = RunIn(file_id=file.id, run_id=context.run.id)
            # create a separate session under-the-hood for this transaction
            ln_add(run_in)
        else:
            # relationship-based, needs session
            if context.run not in file.input_of:
                file.input_of.append(context.run)
                session = object_session(file)
                session.add(file)
                session.commit()


def load(file: File, is_run_input: Optional[bool] = None) -> DataLike:
    """Stage and load to memory.

    Returns in-memory representation if possible, e.g., an `AnnData` object
    for an `h5ad` file.
    """
    _track_run_input(file, is_run_input)
    return load_to_memory(filepath_from_file(file))


def stage(file: File, is_run_input: Optional[bool] = None) -> Path:
    """Update cache from cloud storage if outdated.

    Returns a path to a locally cached on-disk object (say, a
    `.jpg` file).
    """
    if file.suffix == ".zarr":
        raise RuntimeError("zarr object can't be staged, please use load() or stream()")
    _track_run_input(file, is_run_input)
    return setup_settings.instance.storage.cloud_to_local(filepath_from_file(file))


File.backed = backed
File.stage = stage
File.subsetter = subsetter
File.stream = stream
File.load = load



"""v0.6.0.

Revision ID: 5fa54c55c3bf
Revises: d1b3e5da6391
Create Date: 2022-09-15 16:46:41.338263

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "5fa54c55c3bf"
down_revision = "d1b3e5da6391"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.drop_table("alembic_version")
    op.add_column(
        "version_yvzi",
        sa.Column("migration", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    )


def downgrade() -> None:
    pass



from typing import Any, List

import pandas as pd
from lndb_storage import load_to_memory
from lnschema_core import Features

from lamindb.dev.db._add import add
from lamindb.dev.db._select import select
from lamindb.dev.hashing import hash_set
from lamindb.schema._table import table_meta


def get_features_records(
    parsing_id: str,
    features_ref: Any,
    df_curated: pd.DataFrame,
) -> List:
    import lnschema_bionty as bionty

    # insert species entry if not exists
    species = select(bionty.Species, name=features_ref.species).one_or_none()
    if species is None:
        species = add(bionty.Species(name=features_ref.species))

    model = table_meta.get_model(f"bionty.{features_ref._entity}")

    # all existing feature records of the species in the db
    stmt = (
        select(model)
        .where(getattr(model, parsing_id).in_(df_curated.index))
        .where(getattr(model, "species_id") == species.id)  # type:ignore
    )
    records = stmt.all()
    records_df = df_curated.index.intersection(stmt.df()[parsing_id])

    # new records to be appended
    new_ids = df_curated.index.difference(records_df)
    if len(new_ids) > 0:
        # mapped new_ids
        reference_df = features_ref.df().set_index(features_ref.reference_id)
        mapped = reference_df.loc[reference_df.index.intersection(new_ids)].copy()
        mapped.index.name = parsing_id
        if mapped.shape[0] > 0:
            for kwargs in mapped.reset_index().to_dict(orient="records"):
                kwargs["species_id"] = species.id  # type:ignore
                record = model(**kwargs)
                records.append(record)
        # unmapped new_ids
        unmapped = set(new_ids).difference(mapped.index)
        if len(unmapped) > 0:
            for i in unmapped:
                record = model(
                    **{parsing_id: i, "species_id": species.id}  # type:ignore
                )
                records.append(record)
    return records


def parse_features(df: pd.DataFrame, features_ref: Any, **curate_kwargs) -> None:
    """Link features to a knowledge table.

    Args:
        df: a DataFrame
        features_ref: Features reference class.
    """
    from bionty import CellMarker, Gene, Protein

    df_curated = features_ref.curate(df=df, **curate_kwargs)
    # ._parsing_id only exist after curate is called
    parsing_id = features_ref._parsing_id

    # logging of curation
    n = df_curated["__curated__"].count()
    n_mapped = df_curated["__curated__"].sum()
    log = {  # noqa  TODO: store this somewhere in the db
        "feature": parsing_id,
        "n_mapped": n_mapped,
        "percent_mapped": round(n_mapped / n * 100, 1),
        "unmapped": df_curated.index[~df_curated["__curated__"]],
    }

    features_hash = hash_set(set(df_curated.index))

    features_type = features_ref._entity

    features = select(
        Features,
        id=features_hash,
        type=features_type,
    ).one_or_none()
    if features is not None:
        return features  # features already exists!
    else:
        features = Features(id=features_hash, type=features_type)
        records = get_features_records(parsing_id, features_ref, df_curated)

        if isinstance(features_ref, Gene):
            for record in records:
                features.genes.append(record)
        elif isinstance(features_ref, Protein):
            for record in records:
                features.proteins.append(record)
        elif isinstance(features_ref, CellMarker):
            for record in records:
                features.cell_markers.append(record)

    return features


def get_features(file_privates, features_ref, **curate_kwargs):
    """Updates file in place."""
    memory_rep = file_privates["_memory_rep"]
    if memory_rep is None:
        memory_rep = load_to_memory(file_privates["_local_filepath"])
    try:
        df = getattr(memory_rep, "var")  # for AnnData objects
        if callable(df):
            df = memory_rep
    except AttributeError:
        df = memory_rep
    return parse_features(df, features_ref, **curate_kwargs)



from lamin_logger import logger
from lndb import __version__ as lndb_v
from lndb_storage import __version__ as lndb_storage_v
from lnschema_core import __version__ as lnschema_core_v
from nbproject import __version__ as nbproject_v
from packaging import version

# Lamin PINNED packages

if version.parse(lnschema_core_v) != version.parse("0.33.4"):
    raise RuntimeError("lamindb needs lnschema_core==0.33.4")

if version.parse(lndb_storage_v) != version.parse("0.2rc6"):
    raise RuntimeError("lamindb needs lndb_storage==0.2rc6")

if version.parse(lndb_v) < version.parse("0.44.7"):
    raise RuntimeError("Upgrade lndb! pip install lndb==0.44.7")

# Lamin GREATEREQ packages

if version.parse(nbproject_v) < version.parse("0.8.5"):
    raise RuntimeError("lamindb needs nbproject>=0.8.5")

# ensure that the lamin package is not installed
try:
    import lamin  # noqa

    logger.warning(
        "Please,\n"
        " - replace `import lamin` with `import lamindb.setup as lnsetup`\n"
        " - run `pip uninstall lamin`\n"
        "lamindb.setup now has all of the lamin functionality\n"
        "The lamindb API and lamin API will be integrated soon!\n"
        "The CLI remains as is!"
    )
except ImportError:
    pass



import nox
import requests  # type: ignore
from laminci import move_built_docs_to_docs_slash_project_slug, upload_docs_artifact
from laminci.nox import build_docs, login_testuser1, run_pre_commit, run_pytest

nox.options.reuse_existing_virtualenvs = True


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def build(session):
    login_testuser1(session)
    session.install(".[dev,test]")
    response = requests.get("https://github.com/laminlabs/lamindb/tree/cloud")
    if response.status_code < 400:
        session.install("git+https://github.com/laminlabs/lamindb@cloud")
    else:
        session.install("git+https://github.com/laminlabs/lamindb")
    run_pytest(session)
    build_docs(session)
    upload_docs_artifact()
    move_built_docs_to_docs_slash_project_slug()



"""v0.26.1.

Revision ID: ff3b5b3ec913
Revises: 8bf788467d0a
Create Date: 2023-02-10 11:27:41.135864

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ff3b5b3ec913"
down_revision = "8bf788467d0a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        op.rename_table("storage", "core.storage")
    else:
        op.execute("alter table public.storage set schema core")


def downgrade() -> None:
    pass



"""IDs.

See: https://github.com/laminlabs/lamin-notes/blob/main/docs/2022/ids.ipynb

Base generators:

.. autosummary::
   :toctree: .

   base26
   base62
   base64

Entity-related generators:

.. autosummary::
   :toctree: .

   schema
   user
   instance
   storage
   file
   folder
   run
   transform
   project
   secret
"""
from ._id import (  # noqa
    base26,
    base62,
    base64,
    file,
    folder,
    instance,
    project,
    run,
    schema,
    secret,
    storage,
    transform,
    user,
)

# backward compat
pipeline = transform
notebook = transform



"""v0.28.0."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "1dafcf0b22aa"
down_revision = "8280855a5064"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    op.alter_column(f"{prefix}dobject", column_name="run_id", new_column_name="source_id", schema=schema)
    if sqlite:
        with op.batch_alter_table("core.dobject", schema=None) as batch_op:
            # batch_op.drop_index("ix_core.dobject_run_id")  # this errors probably on old instances
            batch_op.create_index(batch_op.f("ix_core.dobject_source_id"), ["source_id"], unique=False)
            batch_op.create_foreign_key(batch_op.f("fk_core.dobject_source_id_run"), "core.run", ["source_id"], ["id"])
    else:
        op.drop_index("ix_core_dobject_run_id", table_name="dobject", schema="core")
        op.create_index(op.f("ix_core_dobject_source_id"), "dobject", ["source_id"], unique=False, schema="core")


def downgrade() -> None:
    pass



from ._file import store_object


def store_png(filepath: str, filekey: str):
    """Store png file."""
    if not filepath.endswith("png"):
        raise ValueError()
    store_object(filepath, filekey)



import random
import string

import pandas as pd
import sqlalchemy as sql


def get_engine():
    return sql.create_engine("sqlite:///test.db", future=True)


def uid(n_char=4):
    base62 = string.digits + string.ascii_letters.swapcase()
    id = "".join(random.choice(base62) for i in range(n_char))
    return id


def create_with_sqlcore():
    metadata = sql.MetaData()

    user = sql.Table(
        "user",
        metadata,
        sql.Column("id", sql.String, primary_key=True, default=uid),
        sql.Column("name", sql.String),
    )

    source = sql.Table(
        "source",
        metadata,
        sql.Column("id", sql.String, primary_key=True, default=uid),
        sql.Column("name", sql.String),
        sql.Column("user", sql.String, sql.ForeignKey("user.id")),
    )

    file = sql.Table(
        "file",
        metadata,
        sql.Column("id", sql.String, primary_key=True, default=uid),
        sql.Column("name", sql.String),
        sql.Column("source", sql.ForeignKey("source.id")),
    )

    engine = get_engine()
    metadata.create_all(bind=engine)

    # insert data
    with engine.begin() as conn:
        stmt = sql.insert(user).values(
            name="falexwolf",
        )
        result = conn.execute(stmt)
        stmt = sql.insert(source).values(
            name="Ingest Alpert19",
            user=result.inserted_primary_key[0],
        )
        result = conn.execute(stmt)
        stmt = sql.insert(file).values(
            name="test_file.csv",
            source=result.inserted_primary_key[0],
        )
        conn.execute(stmt)


def load_table_pandas(table_name):
    engine = get_engine()
    with engine.connect() as conn:
        df = pd.read_sql_table(table_name, conn, index_col="id")
    return df


def test_sqlcore_to_pandas():
    create_with_sqlcore()
    load_table_pandas("user")
    load_table_pandas("source")
    load_table_pandas("file")


if __name__ == "__main__":
    test_sqlcore_to_pandas()



import importlib
from typing import Optional

from IPython.display import display
from lamin_logger import colors
from lndb import settings
from lndb.dev._setup_schema import get_schema_module_name

from .dev.db._select import select


def view(n: int = 10, schema: Optional[str] = None):
    """View data.

    Args:
        n: Display the latest n rows of the tables.
        schema: Schema module to view. Default's to `None` and
            displays all schema modules.
    """
    if schema is not None:
        schema_names = [schema]
    else:
        schema_names = ["core"] + list(settings.instance.schema)

    for schema_name in schema_names:
        schema_module = importlib.import_module(get_schema_module_name(schema_name))

        tables = [
            i
            for i in schema_module.__dict__.values()
            if i.__class__.__name__ == "SQLModelMetaclass" and hasattr(i, "__table__")
        ]
        section = f"* module: {colors.green(colors.bold(schema_name))} *"
        section_no_color = f"* module: {schema_name} *"
        print("*" * len(section_no_color))
        print(section)
        print("*" * len(section_no_color))
        for entity in tables:
            df = select(entity).df()
            if df.shape[0] > 0:
                print(colors.blue(colors.bold(entity.__name__)))
                display(df.iloc[-n:])



"""v0.30rc2."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op
from sqlalchemy.engine.reflection import Inspector

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic
from lnschema_core.dev.type import TransformType

revision = "ebafd37fd6e1"
down_revision = "9640062eefee"


copy_pipeline_to_notebook = """\
insert into {core_notebook} (id, v, name, reference, created_by, created_at, updated_at)
select id, v, name, reference, created_by, created_at, updated_at from {core_pipeline}
"""

make_pipeline_references_notebook_references_in_run = """\
update {core_run}
set notebook_id = pipeline_id, notebook_v = pipeline_v
where notebook_id is null
"""

SATransformType = sa.Enum(TransformType)


def upgrade() -> None:
    sqlite, prefix, schema, delim = get_sqlite_prefix_schema_delim_from_alembic()

    engine = op.get_bind().engine
    inspector = Inspector.from_engine(engine)

    op.add_column(f"{prefix}notebook", sa.Column("reference", sqm.sql.sqltypes.AutoString(), nullable=True), schema=schema)

    core_notebook = "core.notebook"
    core_pipeline = "core.pipeline"
    core_run = "core.run"
    if sqlite:  # sqlite needs quotes
        core_notebook, core_pipeline, core_run = f"'{core_notebook}'", f"'{core_pipeline}'", f"'{core_run}'"

    # add the type column to the existing notebook table
    SATransformType.create(op.get_bind(), checkfirst=True)
    op.add_column(f"{prefix}notebook", sa.Column("type", SATransformType), schema=schema)

    op.execute(f"update {core_notebook} set type = 'notebook'")
    op.execute(copy_pipeline_to_notebook.format(core_notebook=core_notebook, core_pipeline=core_pipeline))
    op.execute(f"update {core_notebook} set type = 'pipeline' where type is null")
    op.execute(make_pipeline_references_notebook_references_in_run.format(core_run=core_run))

    op.drop_index(f"ix_core{delim}pipeline_created_at", table_name=f"{prefix}pipeline", schema=schema)
    op.drop_index(f"ix_core{delim}pipeline_created_by", table_name=f"{prefix}pipeline", schema=schema)
    op.drop_index(f"ix_core{delim}pipeline_name", table_name=f"{prefix}pipeline", schema=schema)
    op.drop_index(f"ix_core{delim}pipeline_reference", table_name=f"{prefix}pipeline", schema=schema)
    op.drop_index(f"ix_core{delim}pipeline_updated_at", table_name=f"{prefix}pipeline", schema=schema)

    op.drop_index(f"ix_core{delim}run_pipeline_id", table_name=f"{prefix}run", schema=schema)
    op.drop_index(f"ix_core{delim}run_pipeline_v", table_name=f"{prefix}run", schema=schema)
    op.drop_index(f"ix_core{delim}run_notebook_id", table_name=f"{prefix}run", schema=schema)
    op.drop_index(f"ix_core{delim}run_notebook_v", table_name=f"{prefix}run", schema=schema)

    with op.batch_alter_table(f"{prefix}run", schema=schema) as batch_op:
        for constraint in inspector.get_foreign_keys(f"{prefix}run", schema=schema):
            if constraint["name"] == "fk_run_notebook_id_notebook":
                batch_op.drop_constraint("fk_run_notebook_id_notebook", type_="foreignkey")
            if constraint["name"] == "fk_run_pipeline_id_pipeline":
                batch_op.drop_constraint("fk_run_pipeline_id_pipeline", type_="foreignkey")

        batch_op.drop_column(column_name="pipeline_id")
        batch_op.drop_column(column_name="pipeline_v")

    op.drop_table(table_name=f"{prefix}pipeline", schema=schema)
    op.rename_table(old_table_name=f"{prefix}notebook", new_table_name=f"{prefix}transform", schema=schema)

    with op.batch_alter_table(f"{prefix}run", schema=schema) as batch_op:
        batch_op.alter_column(column_name="notebook_id", new_column_name="transform_id")
        batch_op.alter_column(column_name="notebook_v", new_column_name="transform_v")

    # need another batch_op as the rename is not yet complete up there
    with op.batch_alter_table(f"{prefix}run", schema=schema) as batch_op:
        batch_op.create_index(op.f("ix_core_run_transform_id"), ["transform_id"], unique=False)
        batch_op.create_index(op.f("ix_core_run_transform_v"), ["transform_v"], unique=False)

        batch_op.create_foreign_key(op.f("fk_run_transform_id_transform"), f"{prefix}transform", ["transform_id", "transform_v"], ["id", "v"], referent_schema=schema)

    with op.batch_alter_table(f"{prefix}transform", schema=schema) as batch_op:
        batch_op.drop_index(f"ix_core{delim}notebook_created_at")
        batch_op.drop_index(f"ix_core{delim}notebook_created_by")
        batch_op.drop_index(f"ix_core{delim}notebook_name")
        batch_op.drop_index(f"ix_core{delim}notebook_title")
        batch_op.drop_index(f"ix_core{delim}notebook_updated_at")
        batch_op.create_index(op.f(f"ix_core{delim}transform_created_at"), ["created_at"], unique=False)
        batch_op.create_index(op.f(f"ix_core{delim}transform_created_by"), ["created_by"], unique=False)
        batch_op.create_index(op.f(f"ix_core{delim}transform_name"), ["name"], unique=False)
        batch_op.create_index(op.f(f"ix_core{delim}transform_title"), ["title"], unique=False)
        batch_op.create_index(op.f(f"ix_core{delim}transform_type"), ["type"], unique=False)
        batch_op.create_index(op.f(f"ix_core{delim}transform_updated_at"), ["updated_at"], unique=False)
        batch_op.alter_column("type", nullable=False)


def downgrade() -> None:
    pass



"""v0.32.1."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic
from lnschema_core.dev.type import TransformType

revision = "d161a14a12e3"
down_revision = "6de59093e378"

# this type got added in v0.30rc2, now we're updating it
SATransformType = sa.Enum(TransformType)


# see https://medium.com/makimo-tech-blog/upgrading-postgresqls-enum-type-with-sqlalchemy-using-alembic-migration-881af1e30abe
def upgrade() -> None:
    sqlite, _, _, _ = get_sqlite_prefix_schema_delim_from_alembic()
    if not sqlite:
        with op.get_context().autocommit_block():
            op.execute("ALTER TYPE transformtype ADD VALUE 'app'")


def downgrade() -> None:
    pass



# it'd be nice to use properties, but class properties are
# a difficult thing
# all hacks via a global instance or a metaclass lead to new hacks
# in the docs, so, we'll go with the below
class settings:
    """Settings.

    This is a static class to manage post-setup settings.

    See `lamindb.settings <https://lamin.ai/docs/db/lamindb.settings>`__.

    For setup-related settings, see
    `lndb.settings <https://lamin.ai/docs/lndb/lndb.settings>`__.
    """

    error_on_file_hash_exists: bool = True
    """Upon ingestion, error if a file hash equals an existing hash in the DB.

    FAQ: :doc:`/faq/ingest-same-file-twice`.
    """
    track_run_inputs_upon_load: bool = False
    """Upon load, add loaded files as the input of the current notebook run.

    FAQ: :doc:`/faq/track-runin`.
    """



import os
import shutil
from pathlib import Path
from typing import Union

import fsspec
import pandas as pd
import readfcs
from lndb import settings
from lndb.dev.upath import UPath

from ._h5ad import read_adata_h5ad
from ._zarr import read_adata_zarr

READER_FUNCS = {
    ".csv": pd.read_csv,
    ".h5ad": read_adata_h5ad,
    ".parquet": pd.read_parquet,
    ".fcs": readfcs.read,
    ".zarr": read_adata_zarr,
}


def print_hook(size, value, **kwargs):
    progress = value / size
    out = f"... uploading {Path(kwargs['filepath']).name}: {min(progress, 1.):4.2f}"
    if progress >= 1:
        out += "\n"
    if "NBPRJ_TEST_NBPATH" not in os.environ:
        print(out, end="\r")


class ProgressCallback(fsspec.callbacks.Callback):
    def branch(self, path_1, path_2, kwargs):
        kwargs["callback"] = fsspec.callbacks.Callback(
            hooks=dict(print_hook=print_hook), filepath=path_1
        )

    def call(self, *args, **kwargs):
        return None


def store_object(localpath: Union[str, Path], storagekey: str) -> float:
    """Store arbitrary file to configured storage location.

    Returns size in bytes.
    """
    storagepath = settings.instance.storage.key_to_filepath(storagekey)
    localpath = Path(localpath)

    if localpath.is_file():
        size = localpath.stat().st_size
    else:
        size = sum(f.stat().st_size for f in localpath.rglob("*") if f.is_file())

    if isinstance(storagepath, UPath):
        if localpath.suffix != ".zarr":
            cb = ProgressCallback()
        else:
            # todo: make proper progress bar for zarr
            cb = fsspec.callbacks.NoOpCallback()
        storagepath.upload_from(localpath, recursive=True, callback=cb)
    else:  # storage path is local
        storagepath.parent.mkdir(parents=True, exist_ok=True)
        if localpath.is_file():
            try:
                shutil.copyfile(localpath, storagepath)
            except shutil.SameFileError:
                pass
        else:
            if storagepath.exists():
                shutil.rmtree(storagepath)
            shutil.copytree(localpath, storagepath)
    return float(size)  # because this is how we store in the db


def delete_storage(storagekey: str):
    """Delete arbitrary file."""
    storagepath = settings.instance.storage.key_to_filepath(storagekey)
    if storagepath.is_file():
        storagepath.unlink()
    elif storagepath.is_dir():
        if isinstance(storagepath, UPath):
            storagepath.rmdir()
        else:
            shutil.rmtree(storagepath)
    else:
        raise FileNotFoundError(f"{storagepath} is not an existing path!")


def load_to_memory(filepath: Union[str, Path, UPath], stream: bool = False):
    """Load a file into memory.

    Returns the filepath if no in-memory form is found.
    """
    if isinstance(filepath, str):
        filepath = Path(filepath)

    if filepath.suffix == ".zarr":
        stream = True
    elif filepath.suffix != ".h5ad":
        stream = False

    if not stream:
        # caching happens here if filename is a UPath
        # todo: make it safe when filepath is just Path
        filepath = settings.instance.storage.cloud_to_local(filepath)

    reader = READER_FUNCS.get(filepath.suffix)
    if reader is None:
        return filepath
    else:
        return reader(filepath)



from typing import Dict, Optional, Tuple, Union

import sqlmodel as sqm
from lnschema_core import File
from sqlmodel import Session


def file_to_sqm(entity: Union[sqm.SQLModel, Tuple[sqm.SQLModel]]):
    def if_file(entity):
        if entity.__class__.__name__ == "type" and entity.__name__ == "File":
            return File
        else:
            return entity

    if isinstance(entity, tuple):
        entities = list(entity)
        for i, ent in enumerate(entities):
            entities[i] = if_file(ent)
        return entities
    else:
        return if_file(entity)


def get_session_from_kwargs(kwargs: Dict) -> Optional[Session]:
    # modifies kwargs inplace if they contain a session object
    if "session" in kwargs:
        session = kwargs.pop("session")
        assert isinstance(session, Session)
        return session
    else:
        return None



from pathlib import Path
from typing import Optional, Union

from anndata import AnnData
from pandas import DataFrame


def infer_suffix(dmem, adata_format: Optional[str] = None):
    """Infer LaminDB storage file suffix from a data object."""
    if isinstance(dmem, AnnData):
        if adata_format is not None:
            if adata_format not in ("h5ad", "zarr"):
                raise ValueError
            return "." + adata_format
        return ".h5ad"
    elif isinstance(dmem, DataFrame):
        return ".parquet"
    else:
        raise NotImplementedError


def write_to_file(dmem, filepath: Union[str, Path]):
    if isinstance(dmem, AnnData):
        dmem.write(filepath)
    elif isinstance(dmem, DataFrame):
        try:
            dmem.to_parquet(filepath)
        except ValueError:
            dmem.reset_index().to_parquet(filepath)
    else:
        raise NotImplementedError



"""Data lineage (`yvzi`)."""
_schema_id = "yvzi"
_name = "core"
_migration = "6a73c00555b4"
__version__ = "0.33.4"

from lndb._check_instance_setup import check_instance_setup as _check_instance_setup

_INSTANCE_SETUP = _check_instance_setup()

if _INSTANCE_SETUP:
    from . import dev, link
    from ._core import Features, File, Folder, Project, Run, Storage, Transform, User

    # backward compat
    DObject = File
    DFolder = Folder



from datetime import datetime as datetime
from typing import Optional

from sqlmodel import Field, SQLModel

from .._timestamps import CreatedAt
from .._users import CreatedBy


class version_yvzi(SQLModel, table=True):  # type: ignore
    """Core schema module versions deployed in a given instance.

    Migrations of the schema module add rows to this table, storing the schema
    module version to which we migrated along with the user who performed the
    migration.
    """

    v: Optional[str] = Field(primary_key=True)
    """Python package version of `lnschema_core`."""
    migration: Optional[str] = None
    """Migration script reference of the latest migration leading up to the Python package version."""  # noqa
    user_id: str = CreatedBy
    created_at: datetime = CreatedAt


class migration_yvzi(SQLModel, table=True):  # type: ignore
    """Latest migration.

    This stores the reference to the latest migration script deployed.
    """

    version_num: Optional[str] = Field(primary_key=True)
    """Reference to the last-run migration script."""



from sqlmodel import Field

from lnschema_core.dev.sqlmodel import schema_sqlmodel

from . import _name as schema_name

SQLModel, prefix, schema_arg = schema_sqlmodel(schema_name)


class RunIn(SQLModel, table=True):  # type: ignore
    """Inputs of runs.

    This is a many-to-many link table for `run` and `file` storing the
    inputs of data transformations.

    A data transformation can have an arbitrary number of data objects as inputs.

    - The same `file` can be used as input in many different `runs`.
    - One `run` can have several `files` as inputs.
    """

    __tablename__ = f"{prefix}run_in"

    run_id: str = Field(foreign_key="core.run.id", primary_key=True)
    file_id: str = Field(foreign_key="core.file.id", primary_key=True)


class FileFeatures(SQLModel, table=True):  # type: ignore
    """Links `File` and `Features`."""

    __tablename__ = f"{prefix}file_features"

    file_id: str = Field(foreign_key="core.file.id", primary_key=True)
    features_id: str = Field(foreign_key="core.features.id", primary_key=True)


class ProjectFolder(SQLModel, table=True):  # type: ignore
    """Link table of project and folder."""

    __tablename__ = f"{prefix}project_folder"

    project_id: str = Field(foreign_key="core.project.id", primary_key=True)
    folder_id: str = Field(foreign_key="core.folder.id", primary_key=True)


class FolderFile(SQLModel, table=True):  # type: ignore
    """Link table of folder and file."""

    __tablename__ = f"{prefix}folder_file"

    folder_id: str = Field(foreign_key="core.folder.id", primary_key=True)
    file_id: str = Field(foreign_key="core.file.id", primary_key=True)



from sqlmodel import Field


def current_user_id():
    from lndb import settings

    return settings.user.id


CreatedBy = Field(default_factory=current_user_id, foreign_key="core.user.id", index=True)



from typing import Dict, List, Union

import pandas as pd
import sqlalchemy as sql
import sqlmodel as sqm


def parse_and_insert_df(df: pd.DataFrame, target_table: str) -> Dict[str, Dict]:
    added_entries = {"mappings": {}, "entries": {}}  # type: ignore
    try:
        # insert columns into relevant tables
        mapper = map_cols_to_tables(df, target_table)
        added_entries = _add_columns(df, mapper)
        df = _populate_cols_with_pks(df, added_entries)
        # insert dataframe into sample table
        sample_sqa = get_sample_table(target_table)
        sample_sqm = get_sqm_tables(sample_sqa.name)
        df_entries = add_from_df(df, sample_sqm)
        added_entries = _update_added_entries(added_entries, df_entries)
    except Exception as e:
        _delete_added_entries(added_entries)
        raise e
    return added_entries


def add_from_column(
    pd_series: pd.Series, table: sqm.SQLModel, table_column: str
) -> Dict[str, Dict]:
    """Quick-insert entries into LaminDB from a DataFrame column.

    This only writes one single column of each entry into the database.
    """
    import lamindb as ln

    entries = pd_series.unique().tolist()
    table_name = str(table.__table__.name)
    added_entries = {"mappings": {}, "entries": {}}  # type: ignore
    added_entries["mappings"][pd_series.name] = (table_name, table_column)
    added_entries["entries"][table_name] = []
    for value in entries:
        if value is None:
            continue
        try:
            entry_data = {table_column: value}
            db_entry = ln.add(table, **entry_data)
            added_entries["entries"][table_name] += [db_entry]
        except Exception as e:
            _delete_added_entries(added_entries)
            raise e
    return added_entries


def add_from_df(
    df: pd.DataFrame, table: sqm.SQLModel, column_map: dict = None
) -> Dict[str, Dict]:
    """Quick-insert all rows of a DataFrame as entries into LaminDB."""
    import lamindb as ln

    added_entries = {"mappings": {}, "entries": {}}  # type: ignore

    # Rename columns
    df_temp = df.copy()
    if column_map:
        df_temp = df_temp.rename(columns=column_map)
    df_temp = df_header_to_snake(df_temp)
    df_temp = df_temp.fillna("None")

    # Get the intersection between df and table columns
    common_fields = set(table.__fields__.keys()).intersection(df_temp.columns)
    common_df_as_records = df_temp[list(common_fields)].to_dict(orient="records")

    # Add df rows to db
    entries = []
    for row in common_df_as_records:
        try:
            entries += [ln.add(table, **row)]
        except Exception as e:
            _delete_added_entries(added_entries)
            raise e

    # format return mappings and entries
    table_name = str(table.__table__.name)
    for field in common_fields:
        try:
            df_col = match_col_from_df(df, field)
            added_entries["mappings"][df_col] = (table_name, field)
        except KeyError:
            # ignore KeyError for intermediate field (pks with fk constraints)
            if field == "id":
                pass
    added_entries["entries"][table_name] = entries

    return added_entries


def str_to_snake(string: str) -> str:
    import re

    string = string.lower()
    string = re.sub("[ ]", "_", string)
    return string


def df_header_to_snake(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    df.columns = [str_to_snake(x) for x in df.columns]
    return df


def get_sample_table(sample: str) -> sql.schema.Table:
    """Get a single sqlalchemy table for the sample of interest."""
    tables = get_sql_tables(sample)
    if not tables:
        raise ValueError(f"Table {sample} does not exist.")
    elif len(tables) == 1:
        return tables[0]
    for table in tables:
        if "wetlab" not in table.name:
            return table


def get_sql_tables(table_name: str) -> List[sql.schema.Table]:
    """Get LaminDB sqlalchemy tables whose name perfectly match the input."""
    import lamindb as ln

    tables = ln.schema._core.get_db_metadata().sorted_tables
    matched = [tab for tab in tables if tab.name.split(".")[-1] == table_name]
    return matched


def get_sql_fk_tables(table: sql.schema.Table) -> List[sql.schema.Table]:
    """Get all tables associated with the input table (by fk constraint)."""
    return [fk.column.table for fk in table.foreign_keys]


def get_sqm_tables(name: str) -> sqm.main.SQLModelMetaclass:
    from functools import reduce
    from inspect import getmembers, isclass

    import lamindb as ln

    name_components = name.split(".")
    module = reduce(getattr, name_components[:-1], ln.schema)
    members = getmembers(module, isclass)
    for member in members:
        if name_components[-1] == _camel_to_snake(member[0]):
            return member[1]


def get_pk_fk(table: sqm.SQLModel) -> Union[None, str]:
    """Gets the foreign key field of the input table's primary key."""
    pk = table.__table__.primary_key.columns.items()
    if len(pk) > 1:
        raise NotImplementedError(
            "Fetching of composite primary key's foreign key not yet implemented."
        )
    pk_fks = pk[0][1].foreign_keys
    if pk_fks:
        pk_fks = list(pk_fks)
        if len(pk_fks) > 1:
            raise NotImplementedError(
                "Composite primary key linking not yet implemented."
            )
        pk_fk = str(pk_fks[0].target_fullname)
        return pk_fk
    else:
        return None


def map_cols_to_tables(
    df: pd.DataFrame, target_table: str
) -> Dict[str, sql.schema.Table]:
    """Map df columns to LaminDB tables associated with columns from input table."""
    mapping = {}
    temp_df = df_header_to_snake(df)
    sample_table = get_sample_table(target_table)
    sample_fk_tables = get_sql_fk_tables(sample_table)
    for column in temp_df.columns:
        for table in sample_fk_tables:
            if column in table.name:
                column = match_col_from_df(df, column)
                mapping[column] = table
    return mapping


def match_col_from_df(df: pd.DataFrame, target: str) -> str:
    """Matches dataframe column to target based on a perfect snake case string match."""
    snake_target = str_to_snake(target)
    snake_df = df_header_to_snake(df)
    try:
        match_index = snake_df.columns.get_loc(snake_target)
    except KeyError:
        raise KeyError(
            f"Could not match {target} to DataFrame columns {list(df.columns)}"
        )
    return df.columns[match_index]


def match_col_from_table(table: sql.schema.Table, substring: str) -> str:
    """Returns column name that contains the input substring."""
    table_cols = table.columns.keys()
    matched_columns = [str(s) for s in table_cols if substring in s]
    return matched_columns[0]


def _add_columns(df: pd.DataFrame, column_mapper: dict) -> Dict[str, Dict]:
    """Add mapped columns from DataFrame into LaminDB."""
    added_entries = {"mappings": {}, "entries": {}}  # type: ignore
    try:
        match_field = "name"
        for column, table in column_mapper.items():
            table_column = match_col_from_table(table, match_field)
            sqm_table = get_sqm_tables(table.name)
            entries = add_from_column(df[column], sqm_table, table_column)
            added_entries = _update_added_entries(added_entries, entries)
    except Exception as e:
        _delete_added_entries(added_entries)
        raise e
    return added_entries


def _camel_to_snake(string: str) -> str:
    import re

    return re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()


def _delete_added_entries(added_entries: dict):
    import lamindb as ln

    for table_entries in added_entries["entries"].values():
        for entry in table_entries:
            ln.delete(entry)


def _populate_cols_with_pks(df, col_entries: dict) -> pd.DataFrame:
    """Populate and rename df columns based on column mappings and DB entries."""
    df = df.copy()
    for col, mapping in col_entries["mappings"].items():
        table, field = mapping
        entries = col_entries["entries"][table]
        id_mapper = {getattr(entry, field): entry.id for entry in entries}
        # handle possible SQLModel casting
        for value in df[col].unique().tolist():
            if str(value) in id_mapper.keys():
                id_mapper[value] = id_mapper[str(value)]
        df[col] = df[col].map(id_mapper)
        df = df.rename(columns={col: f"{col} ID"})
    return df


def _update_added_entries(added_entries: Dict, new_entries: Dict) -> Dict:
    added_entries["mappings"] = {**added_entries["mappings"], **new_entries["mappings"]}
    added_entries["entries"] = {**added_entries["entries"], **new_entries["entries"]}
    return added_entries



"""v0.7.1.

Revision ID: 049d7dfc80a8
Revises: 3b60b87450c0
Create Date: 2022-09-18 17:31:43.384602
"""
import sqlalchemy as sa
import sqlmodel  # noqa
from alembic import op

revision = "049d7dfc80a8"
down_revision = "3b60b87450c0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        op.add_column(
            "dobject",
            sa.Column("suffix", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        )
        op.create_index(op.f("ix_dobject_suffix"), "dobject", ["suffix"], unique=False)
        op.execute("update dobject set suffix = file_suffix")
        op.drop_index("ix_dobject_file_suffix", table_name="dobject")
        op.drop_column("dobject", "file_suffix")
    else:
        op.alter_column(
            "dobject",
            column_name="file_suffix",
            new_column_name="suffix",
            nullable=True,
        )


def downgrade() -> None:
    pass



"""v0.4.0.

Revision ID: 01fcb82dafd4
Revises: 1c531ea346cf
Create Date: 2022-08-22 16:01:42.592859

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "01fcb82dafd4"
down_revision = "1c531ea346cf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Create the new pipeline_run table
    op.create_table(
        "pipeline_run",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.drop_table("pipeline")

    # Update the dtransform table
    op.create_table(
        "dtransform_tmp",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("jupynb_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("jupynb_v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("pipeline_run_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"],
            ["jupynb.id", "jupynb.v"],
            name="dtransform_jupynb",
        ),
        sa.ForeignKeyConstraint(
            ["pipeline_run_id"],
            ["pipeline_run.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute("insert into dtransform_tmp (id, jupynb_id, jupynb_v) select id, jupynb_id, jupynb_v from dtransform")
    op.drop_table("dtransform")
    op.rename_table("dtransform_tmp", "dtransform")
    op.create_index(op.f("ix_dtransform_jupynb_id"), "dtransform", ["jupynb_id"], unique=False)
    op.create_index(op.f("ix_dtransform_jupynb_v"), "dtransform", ["jupynb_v"], unique=False)
    op.create_index(
        op.f("ix_dtransform_pipeline_run_id"),
        "dtransform",
        ["pipeline_run_id"],
        unique=False,
    )

    # Drop the redundant dtransform_out table
    op.drop_table("dtransform_out")

    # Rename the dsource_id column to dtransform_id and add indexes
    op.create_table(
        "dobject_tmp",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True, index=True),
        sa.Column(
            "file_suffix",
            sqlmodel.sql.sqltypes.AutoString(),
            nullable=False,
            index=True,
        ),
        sa.Column("dtransform_id", sqlmodel.sql.sqltypes.AutoString(), index=True),
        sa.Column("storage_root", sqlmodel.sql.sqltypes.AutoString(), index=True),
        sa.Column("time_created", sa.DateTime(), nullable=False, index=True),
        sa.Column("time_updated", sa.DateTime(), nullable=False, index=True),
        sa.ForeignKeyConstraint(
            ["dtransform_id"],
            ["dtransform.id"],
        ),
        sa.ForeignKeyConstraint(
            ["storage_root"],
            ["storage.root"],
        ),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.execute(
        "insert into dobject_tmp(id, v, name, file_suffix, dtransform_id,"
        " time_created, time_updated) select id, v, name, file_suffix, dsource_id,"
        " time_created, time_updated from dobject"
    )
    op.drop_table("dobject")
    op.rename_table("dobject_tmp", "dobject")

    # Create indexes on jupynb
    op.create_index(op.f("ix_jupynb_name"), "jupynb", ["name"], unique=False)
    op.create_index(op.f("ix_jupynb_time_created"), "jupynb", ["time_created"], unique=False)
    op.create_index(op.f("ix_jupynb_time_updated"), "jupynb", ["time_updated"], unique=False)
    op.create_index(op.f("ix_jupynb_user_id"), "jupynb", ["user_id"], unique=False)

    # Create indexes on usage
    op.create_index(op.f("ix_usage_dobject_id"), "usage", ["dobject_id"], unique=False)
    op.create_index(op.f("ix_usage_dobject_v"), "usage", ["dobject_v"], unique=False)
    op.create_index(op.f("ix_usage_time"), "usage", ["time"], unique=False)
    op.create_index(op.f("ix_usage_type"), "usage", ["type"], unique=False)
    op.create_index(op.f("ix_usage_user_id"), "usage", ["user_id"], unique=False)

    # Create indexes on user
    op.create_index(op.f("ix_user_email"), "user", ["email"], unique=False)
    op.create_index(op.f("ix_user_handle"), "user", ["handle"], unique=False)


def downgrade() -> None:
    pass



import base64
from pathlib import Path

from lamindb.dev.hashing import hash_file, to_b64_str


def test_compute_hash():
    dataset = [
        ("file_1.txt", "a", "DMF1ucDxtqgxw5niaXcmYQ"),
    ]
    for path, content, hash in dataset:
        filepath = Path(path)
        with open(path, "w") as file:
            file.write(content)
        assert hash == hash_file(filepath)
        filepath.unlink()


def test_base64():
    # the following can be commented out over time
    mytest = "test".encode()
    b64_str = to_b64_str(mytest)
    b64_str_padded = f"{b64_str}=="
    assert base64.urlsafe_b64decode(b64_str_padded.encode()).hex() == mytest.hex()



"""Dev tools for core schema (`yvzi`).

Tracking versions & migrations:

.. autosummary::
   :toctree: .

   version_yvzi
   migration_yvzi

Auxiliary modules:

.. autosummary::
   :toctree: .

   type
   id
   sqlmodel

"""

from . import id, sqlmodel, type  # noqa
from ._versions import migration_yvzi, version_yvzi  # noqa



"""v0.11.0.

Revision ID: 3d244a8d3148
Revises: 0c819d33ca9b
Create Date: 2022-10-10 11:23:16.199020

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3d244a8d3148"
down_revision = "0c819d33ca9b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.add_column(sa.Column("name", sqlmodel.sql.sqltypes.AutoString()))
        batch_op.create_index(batch_op.f("ix_user_name"), ["name"], unique=False)


def downgrade() -> None:
    pass



"""v0.33.0."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic

revision = "6a73c00555b4"
down_revision = "d161a14a12e3"


def upgrade() -> None:
    sqlite, prefix, schema, delim = get_sqlite_prefix_schema_delim_from_alembic()

    op.alter_column(f"{prefix}file", column_name="source_id", new_column_name="run_id", schema=schema)
    op.drop_index(f"ix_core{delim}file_source_id", table_name=f"{prefix}file", schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_run_id"), f"{prefix}file", ["run_id"], unique=False, schema=schema)

    op.alter_column(f"{prefix}folder", column_name="created_by", new_column_name="created_by_id", schema=schema)
    op.alter_column(f"{prefix}run", column_name="created_by", new_column_name="created_by_id", schema=schema)
    op.alter_column(f"{prefix}features", column_name="created_by", new_column_name="created_by_id", schema=schema)
    op.create_index(op.f(f"ix_core{delim}features_created_by_id"), f"{prefix}features", ["created_by_id"], unique=False, schema=schema)
    op.alter_column(f"{prefix}transform", column_name="created_by", new_column_name="created_by_id", schema=schema)
    op.alter_column(f"{prefix}project", column_name="created_by", new_column_name="created_by_id", schema=schema)
    op.add_column(f"{prefix}file", sa.Column("created_by_id", sqm.sql.sqltypes.AutoString(), nullable=True), schema=schema)
    op.add_column(f"{prefix}file", sa.Column("transform_id", sqm.sql.sqltypes.AutoString(), nullable=True), schema=schema)
    op.add_column(f"{prefix}file", sa.Column("transform_version", sqm.sql.sqltypes.AutoString(), nullable=True), schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_created_by_id"), f"{prefix}file", ["created_by_id"], unique=False, schema=schema)
    op.add_column(f"{prefix}storage", sa.Column("created_by_id", sqm.sql.sqltypes.AutoString(), nullable=True), schema=schema)

    with op.batch_alter_table(f"{prefix}file", schema=schema) as batch_op:
        batch_op.create_foreign_key(op.f(f"fk_file{delim}created_by_id_user"), f"{prefix}user", ["created_by_id"], ["id"], referent_schema=schema)

    op.create_index(op.f(f"ix_core{delim}project_created_by_id"), f"{prefix}project", ["created_by_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}run_created_by_id"), f"{prefix}run", ["created_by_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}transform_created_by_id"), f"{prefix}transform", ["created_by_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}folder_created_by_id"), f"{prefix}folder", ["created_by_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_transform_id"), f"{prefix}file", ["transform_id"], unique=False, schema=schema)

    op.alter_column(f"{prefix}transform", column_name="v", new_column_name="version", schema=schema)
    op.alter_column(f"{prefix}run", column_name="transform_v", new_column_name="transform_version", schema=schema)

    with op.batch_alter_table(f"{prefix}file", schema=schema) as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_file_transform_id_version_transform"), f"{prefix}transform", ["transform_id", "transform_version"], ["id", "version"], referent_schema=schema
        )
        batch_op.alter_column("run_id", existing_type=sqm.sql.sqltypes.AutoString(), nullable=True)

    op.create_index(op.f(f"ix_core{delim}file_transform_version"), f"{prefix}file", ["transform_version"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}run_transform_version"), f"{prefix}run", ["transform_version"], unique=False, schema=schema)

    try:
        op.drop_index(f"ix_core{delim}features_created_by", table_name=f"{prefix}features", schema=schema)
        op.drop_index(f"ix_core{delim}run_transform_v", f"{prefix}run", schema=schema)
        op.drop_index(f"ix_core{delim}project_created_by", table_name=f"{prefix}project", schema=schema)
        op.drop_index(f"ix_core{delim}run_created_by", table_name=f"{prefix}run", schema=schema)
        op.drop_index(f"ix_core{delim}transform_created_by", table_name=f"{prefix}transform", schema=schema)
        op.drop_index(f"ix_core{delim}folder_created_by", table_name=f"{prefix}folder", schema=schema)
    except Exception:
        pass


def downgrade() -> None:
    pass



import scipy.sparse as sparse
from pandas import DataFrame


def _size_val(val):
    if sparse.issparse(val):
        val_csr = sparse.csr_matrix(val)
        return val_csr.data.nbytes + val_csr.indptr.nbytes + val_csr.indices.nbytes
    else:
        return val.__sizeof__()


def _size_elem(elem):
    if hasattr(elem, "keys") and not isinstance(elem, DataFrame):
        return sum([_size_val(elem[k]) for k in elem.keys()])
    else:
        return _size_val(elem)


def _size_raw(raw):
    return _size_val(raw.X) + _size_val(raw.var) + _size_elem(raw.varm)


def size_adata(adata):
    total_size = adata.__sizeof__()

    raw = adata.raw
    if raw is not None:
        total_size += _size_raw(raw)

    return total_size



from enum import Enum
from pathlib import Path
from typing import TypeVar

import anndata as ad
import pandas as pd
from lndb.dev.upath import UPath

PathLike = TypeVar("PathLike", str, Path, UPath)
DataLike = TypeVar("DataLike", ad.AnnData, pd.DataFrame)


class Usage(str, Enum):
    """Data access types."""

    ingest = "ingest"
    insert = "insert"
    select = "select"
    update = "update"
    delete = "delete"
    load = "load"
    link = "link"


class TransformType(Enum):
    pipeline = "pipeline"
    notebook = "notebook"
    app = "app"

    def __repr__(self):
        return self.name



"""v0.15.0.

Revision ID: 98da12fc80a8
Revises: cf5913791674
Create Date: 2022-10-31
"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

revision = "98da12fc80a8"
down_revision = "cf5913791674"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        op.rename_table(old_table_name="dtransform", new_table_name="core.dtransform")
        op.rename_table(old_table_name="storage", new_table_name="core.storage")
        op.rename_table(old_table_name="user", new_table_name="core.user")
        op.rename_table(old_table_name="dtransform_in", new_table_name="core.dtransform_in")
        op.rename_table(old_table_name="dobject", new_table_name="core.dobject")
        op.rename_table(old_table_name="usage", new_table_name="core.usage")
        op.rename_table(old_table_name="jupynb", new_table_name="core.jupynb")
        op.rename_table(old_table_name="pipeline", new_table_name="core.pipeline")
        op.rename_table(old_table_name="pipeline_run", new_table_name="core.pipeline_run")
        op.rename_table(old_table_name="dset", new_table_name="core.dset")
        op.rename_table(old_table_name="project", new_table_name="core.project")
        op.rename_table(old_table_name="project_dset", new_table_name="core.project_dset")
        op.rename_table(old_table_name="dset_dobject", new_table_name="core.dset_dobject")
    else:
        op.execute("alter table public.dtransform set schema core")
        op.execute("alter table public.storage set schema core")
        op.execute("alter table public.user set schema core")
        op.execute("alter table public.dtransform_in set schema core")
        op.execute("alter table public.dobject set schema core")
        op.execute("alter table public.usage set schema core")
        op.execute("alter table public.jupynb set schema core")
        op.execute("alter table public.pipeline set schema core")
        op.execute("alter table public.pipeline_run set schema core")
        op.execute("alter table public.dset set schema core")
        op.execute("alter table public.project set schema core")
        op.execute("alter table public.project_dset set schema core")
        op.execute("alter table public.dset_dobject set schema core")


def downgrade() -> None:
    pass



"""v0.30rc1."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "9640062eefee"
down_revision = "873683a29806"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema, delim = "core.", None, "."
    else:
        prefix, schema, delim = "", "core", "_"

    with op.batch_alter_table(f"{prefix}usage", schema=schema) as batch_op:
        batch_op.drop_index(f"ix_core{delim}usage_dobject_id")
        batch_op.drop_index(f"ix_core{delim}usage_time")
        batch_op.drop_index(f"ix_core{delim}usage_type")
        batch_op.drop_index(f"ix_core{delim}usage_user_id")
    op.drop_table(f"{prefix}usage", schema=schema)


def downgrade() -> None:
    pass



"""Object → storage.

Memory → file:

.. autosummary::
   :toctree: .

   anndata_to_h5ad
"""

from ._anndata import anndata_to_h5ad
from ._anndata_sizes import size_adata
from ._core import infer_suffix, write_to_file
from ._lazy_field import LazySelector, lazy
from ._subset_anndata import _subset_anndata_file



import lnschema_core

from .. import schema


def _list_methods(module):
    return [getattr(module, i) for i in dir(module) if not i.startswith("_")]


class table_meta:
    orm_to_model = {}
    tablename_to_orm = {}
    orm_to_tablename = {}
    for schema_pkg in _list_methods(schema) + [lnschema_core]:
        alltables_pkg = _list_methods(schema_pkg)
        try:
            dev_tables = _list_methods(getattr(schema_pkg, "dev"))
        except AttributeError:
            dev_tables = []
        for table in alltables_pkg + dev_tables:
            if table.__class__.__name__ != "SQLModelMetaclass":
                continue
            if not hasattr(table, "__table__"):
                continue
            orm_to_model[table.__name__] = table
            tablename_to_orm[table.__table__.name] = table.__name__
            orm_to_tablename[table.__name__] = table.__table__.name

    @classmethod
    def convert_to_orm(cls, table_name: str):
        """Convert to camel case class name."""
        return (
            table_name
            if table_name.lower() != table_name
            else cls.tablename_to_orm.get(table_name)
        )

    @classmethod
    def get_model(cls, table_name: str):
        name = cls.convert_to_orm(table_name)
        if name is None:
            raise KeyError(f"Table {table_name} does NOT exist!")
        model = cls.orm_to_model.get(name)
        if model is None:
            raise AssertionError(f"Table {name} does NOT exist!")
        return model



"""v0.10.0.

Revision ID: 0c819d33ca9b
Revises: 439c4ee0a22a
Create Date: 2022-10-07 12:47:10.583291

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0c819d33ca9b"
down_revision = "439c4ee0a22a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("jupynb", sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString()))
    op.execute("update jupynb set created_by = user_id")
    with op.batch_alter_table("jupynb", schema=None) as batch_op:
        batch_op.drop_index("ix_jupynb_user_id")
        batch_op.create_index(batch_op.f("ix_jupynb_created_by"), ["created_by"], unique=False)
        batch_op.create_foreign_key("fk_created_by", "user", ["created_by"], ["id"])
        batch_op.drop_column("user_id")

    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.add_column(sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString()))
        batch_op.create_index(batch_op.f("ix_pipeline_created_by"), ["created_by"], unique=False)
        batch_op.create_foreign_key("fk_created_by", "user", ["created_by"], ["id"])

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString()))
        batch_op.create_index(batch_op.f("ix_pipeline_run_created_by"), ["created_by"], unique=False)
        batch_op.create_foreign_key("fk_created_by", "user", ["created_by"], ["id"])

    with op.batch_alter_table("version_yvzi", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_version_yvzi_user_id"), ["user_id"], unique=False)


def downgrade() -> None:
    pass



"""v0.7.0.

Revision ID: 3b60b87450c0
Revises: 5fa54c55c3bf
Create Date: 2022-09-18 15:57:25.104305
"""
import sqlalchemy as sa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3b60b87450c0"
down_revision = "5fa54c55c3bf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("v", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("dtransform_id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("storage_id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.drop_index("ix_dobject_time_created")
        batch_op.drop_index("ix_dobject_time_updated")
        batch_op.create_index(batch_op.f("ix_dobject_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_dobject_updated_at"), ["updated_at"], unique=False)
        batch_op.create_foreign_key("fk_object_storage_id_storage", "storage", ["storage_id"], ["id"])
        batch_op.drop_column("time_created")
        batch_op.drop_column("time_updated")

    with op.batch_alter_table("dtransform", schema=None) as batch_op:
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)

    with op.batch_alter_table("jupynb", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("v", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.drop_index("ix_jupynb_time_created")
        batch_op.drop_index("ix_jupynb_time_updated")
        batch_op.create_index(batch_op.f("ix_jupynb_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_jupynb_updated_at"), ["updated_at"], unique=False)
        batch_op.drop_column("time_created")
        batch_op.drop_column("time_updated")

    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("v", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_index(batch_op.f("ix_pipeline_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_pipeline_updated_at"), ["updated_at"], unique=False)
        batch_op.drop_column("time_created")

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_index(batch_op.f("ix_pipeline_run_created_at"), ["created_at"], unique=False)
        batch_op.create_foreign_key("pipeline", "pipeline", ["pipeline_id", "pipeline_v"], ["id", "v"])
        batch_op.drop_column("time_created")

    with op.batch_alter_table("storage", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_index(batch_op.f("ix_storage_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_storage_updated_at"), ["updated_at"], unique=False)
        batch_op.drop_column("time_created")
        batch_op.drop_column("time_updated")

    with op.batch_alter_table("usage", schema=None) as batch_op:
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)

    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.alter_column("id", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("handle", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_index(batch_op.f("ix_user_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_user_updated_at"), ["updated_at"], unique=False)
        batch_op.drop_column("time_created")
        batch_op.drop_column("time_updated")

    with op.batch_alter_table("version_yvzi", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=True,
            )
        )
        batch_op.alter_column("v", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_index(batch_op.f("ix_version_yvzi_created_at"), ["created_at"], unique=False)
        batch_op.drop_column("time_created")


def downgrade() -> None:
    pass



"""v0.7.3.

Revision ID: 1f29517759b7
Revises: 049d7dfc80a8
Create Date: 2022-09-24 10:32:37.801549

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1f29517759b7"
down_revision = "049d7dfc80a8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("jupynb", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("storage", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )
        # batch_op.drop_constraint("user", type_="unique")  # seems not needed
        batch_op.drop_index("ix_user_email")
        batch_op.create_index(batch_op.f("ix_user_email"), ["email"], unique=True)
        batch_op.drop_index("ix_user_handle")
        batch_op.create_index(batch_op.f("ix_user_handle"), ["handle"], unique=True)

    with op.batch_alter_table("version_yvzi", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            nullable=False,
            existing_server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )

    with op.batch_alter_table("usage", schema=None) as batch_op:
        batch_op.alter_column(
            "time",
            existing_type=sa.DATETIME(),
            nullable=False,
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
        )


def downgrade() -> None:
    pass



"""Development tools.

Tracking versions & migrations:

.. autosummary::
   :toctree: .

   version_vvhc
   migration_vvhc

Auxiliary modules:

.. autosummary::
   :toctree: .

   id

Datasets:

.. autosummary::
   :toctree: .

   datasets

"""
from . import datasets, id
from ._sample_submission import parse_and_insert_df
from ._versions import migration_vvhc, version_vvhc



import os
import shutil
from pathlib import Path
from time import perf_counter

import nox
from laminci import move_built_docs_to_docs_slash_project_slug, upload_docs_artifact
from laminci.nox import (
    build_docs,
    login_testuser1,
    login_testuser2,
    run_pre_commit,
    run_pytest,
)

import lamindb as ln

nox.options.reuse_existing_virtualenvs = True


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
@nox.parametrize("package", ["lamindb", "lndb-storage"])
def build(session, package):
    t_start = perf_counter()
    login_testuser2(session)
    login_testuser1(session)
    t_total = perf_counter() - t_start
    print(f"Done logging in: {t_total:.3f}s")

    t_start = perf_counter()
    # run with pypi install on main
    if "GITHUB_EVENT_NAME" in os.environ and os.environ["GITHUB_EVENT_NAME"] != "push":
        # run with submodule install on a PR
        session.install("./sub/lnschema-core[dev,test]")
        session.install("./sub/lnbase-biolab[dev,test]")
        session.install("./sub/lndb-storage[dev,test]")

    session.install(".[dev,test]")
    t_total = perf_counter() - t_start
    print(f"Done installing: {t_total:.3f}s")

    if package == "lamindb":
        run_pytest(session)
    else:
        # navigate into submodule so that lamin-project.yml is correctly read
        os.chdir(f"./sub/{package}")
        session.run("pytest", "-s", "./tests", "--ignore", "./tests/test_migrations.py")

    if package == "lamindb":
        t_start = perf_counter()

        # Schemas
        ln.setup.load("testuser1/lamin-site-assets", migrate=True)

        file = ln.select(ln.File, key="docs/lndb_storage_docs.zip").one()
        shutil.unpack_archive(file.stage(), "lndb_storage_docs")
        Path("lndb_storage_docs/guide/stream.ipynb").rename("docs/guide/stream.ipynb")

        file = ln.select(ln.File, key="docs/lnschema_core_docs.zip").one()
        shutil.unpack_archive(file.stage(), "lnschema_core_docs")
        Path("lnschema_core_docs/guide/0-core-schema.ipynb").rename(
            "docs/guide/lnschema-core.ipynb"
        )
        Path("lnschema_core_docs/guide/1-data-validation.ipynb").rename(
            "docs/guide/data-validation.ipynb"
        )

        file = ln.select(ln.File, key="docs/lnschema_bionty_docs.zip").one()
        shutil.unpack_archive(file.stage(), "lnschema_bionty_docs")
        Path("lnschema_bionty_docs/guide/bionty-orms.ipynb").rename(
            "docs/guide/lnschema-bionty.ipynb"
        )
        Path("lnschema_bionty_docs/guide/knowledge.ipynb").rename(
            "docs/guide/knowledge.ipynb"
        )

        t_total = perf_counter() - t_start
        print(f"Done pulling artifacts: {t_total:.3f}s")

        t_start = perf_counter()
        build_docs(session)
        login_testuser1(session)
        upload_docs_artifact()
        move_built_docs_to_docs_slash_project_slug()
        print(f"Done building docs and uploading: {t_total:.3f}s")



"""v0.28.7."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "873683a29806"
down_revision = "3dd9f8d41861"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    with op.batch_alter_table(f"{prefix}dobject", schema=schema) as batch_op:
        batch_op.create_unique_constraint("uq_storage__objectkey_suffix", ["storage_id", "_objectkey", "suffix"])


def downgrade() -> None:
    pass



"""v0.8.0.

Revision ID: 7e8f7b30792e
Revises: 1f29517759b7
Create Date: 2022-09-25 20:39:03.626212
"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7e8f7b30792e"
down_revision = "1f29517759b7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.add_column(sa.Column("size", sa.Float(), nullable=True))
        batch_op.create_index(batch_op.f("ix_dobject_size"), ["size"], unique=False)


def downgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_dobject_size"))
        batch_op.drop_column("size")



"""v0.17.0.

Revision ID: 66bfd6cf2e2d
Revises: 4ee426b656bb
Create Date: 2022-11-11 22:05:39.865585

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "66bfd6cf2e2d"
down_revision = "4ee426b656bb"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
        op.execute("PRAGMA foreign_keys=OFF")
    else:
        prefix, schema = "", "core"

    with op.batch_alter_table(f"{prefix}dtransform", schema=schema) as batch_op:
        batch_op.add_column(sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
        batch_op.add_column(sa.Column("pipeline_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
        batch_op.add_column(sa.Column("pipeline_v", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
        batch_op.create_index(batch_op.f("ix_core.dtransform_pipeline_id"), ["pipeline_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dtransform_pipeline_v"), ["pipeline_v"], unique=False)
        batch_op.add_column(
            sa.Column(
                "created_by",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=False,
                default="9ypQ1yrW",
            )
        )
        batch_op.add_column(
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=False,
            )
        )
        batch_op.create_index(batch_op.f("ix_core.dtransform_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dtransform_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dtransform_name"), ["name"], unique=False)
        if sqlite:
            batch_op.create_foreign_key(
                batch_op.f("fk_core.dtransform_pipeline_id_pipeline"),
                f"{prefix}pipeline",
                ["pipeline_id", "pipeline_v"],
                ["id", "v"],
            )
            batch_op.create_foreign_key(
                batch_op.f("fk_core.dtransform_created_by_user"),
                f"{prefix}user",
                ["created_by"],
                ["id"],
            )

    if not sqlite:
        op.create_foreign_key(
            "fk_core.dtransform_pipeline_id_pipeline",
            f"{prefix}dtransform",
            f"{prefix}pipeline",
            ["pipeline_id", "pipeline_v"],
            ["id", "v"],
            source_schema=schema,
            referent_schema=schema,
        )
        op.create_foreign_key(
            "fk_core.dtransform_created_by_user",
            f"{prefix}dtransform",
            f"{prefix}user",
            ["created_by"],
            ["id"],
            source_schema=schema,
            referent_schema=schema,
        )

    if sqlite:
        op.execute(
            """
        update "core.dtransform"
        set created_by = "core.run".created_by,
        created_at = "core.run".created_at,
        name = "core.run".name,
        pipeline_id = "core.run".pipeline_id,
        pipeline_v = "core.run".pipeline_v
        from "core.run"
        where run_id = "core.run".id;
        """
        )
    else:
        op.execute(
            """
        update core.dtransform
        set created_by = core.run.created_by,
        created_at = core.run.created_at,
        name = core.run.name,
        pipeline_id = core.run.pipeline_id,
        pipeline_v = core.run.pipeline_v
        from core.run
        where run_id = core.run.id;
        """
        )

    try:
        op.drop_index("ix_core.dtransform_run_id")
        op.drop_constraint("bfx_run_id_fkey", "run", schema="bfx")
        op.drop_constraint("dtransform_pipeline_run_id_fkey", "dtransform", schema="core")
    except Exception:
        pass
    op.drop_column(f"{prefix}dtransform", "run_id", schema=schema)
    op.drop_table(f"{prefix}run", schema=schema)

    op.rename_table(
        old_table_name=f"{prefix}dtransform",
        new_table_name=f"{prefix}run",
        schema=schema,
    )
    op.rename_table(
        old_table_name=f"{prefix}dtransform_in",
        new_table_name=f"{prefix}run_in",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}dobject",
        column_name="dtransform_id",
        new_column_name="run_id",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}run_in",
        column_name="dtransform_id",
        new_column_name="run_id",
        schema=schema,
    )

    if bind.engine.name == "sqlite":
        op.execute("PRAGMA foreign_keys=ON")


def downgrade() -> None:
    pass



from sqlalchemy.sql import func
from sqlmodel import Field

CreatedAt = Field(index=True, sa_column_kwargs=dict(server_default=func.now()))
UpdatedAt = Field(default=None, index=True, sa_column_kwargs=dict(onupdate=func.now()))



"""v0.21.1.

Revision ID: 4b4005b7841c
Revises: 66bfd6cf2e2d
Create Date: 2022-11-28 00:38:59.653534

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4b4005b7841c"
down_revision = "66bfd6cf2e2d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
        op.execute("PRAGMA foreign_keys=OFF")
    else:
        prefix, schema = "", "core"

    if sqlite:
        op.rename_table("core.storage", "storage")
        op.create_table(
            "core.features",
            sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column(
                "created_at",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=False,
            ),
            sa.ForeignKeyConstraint(
                ["created_by"],
                ["core.user.id"],
                name=op.f("fk_core.features_created_by_user"),
            ),
            sa.PrimaryKeyConstraint("id", name=op.f("pk_core.features")),
        )
        with op.batch_alter_table("core.features", schema=None) as batch_op:
            batch_op.create_index(batch_op.f("ix_core.features_created_at"), ["created_at"], unique=False)
            batch_op.create_index(batch_op.f("ix_core.features_created_by"), ["created_by"], unique=False)

        op.create_table(
            "core.dobject_features",
            sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("features_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.ForeignKeyConstraint(
                ["dobject_id"],
                ["core.dobject.id"],
                name=op.f("fk_core.dobject_features_dobject_id_dobject"),
            ),
            sa.ForeignKeyConstraint(
                ["features_id"],
                ["core.features.id"],
                name=op.f("fk_core.dobject_features_features_id_features"),
            ),
            sa.PrimaryKeyConstraint("dobject_id", "features_id", name=op.f("pk_core.dobject_features")),
        )
        with op.batch_alter_table("core.storage", schema=None) as batch_op:
            batch_op.drop_index("ix_core.storage_created_at")
            batch_op.drop_index("ix_core.storage_root")
            batch_op.drop_index("ix_core.storage_updated_at")

    op.alter_column(
        f"{prefix}dobject",
        column_name="checksum",
        new_column_name="hash",
        schema=schema,
    )

    with op.batch_alter_table("core.run", schema=None) as batch_op:
        batch_op.drop_index("ix_core.dtransform_created_at")
        batch_op.drop_index("ix_core.dtransform_created_by")
        batch_op.drop_index("ix_core.dtransform_jupynb_id")
        batch_op.drop_index("ix_core.dtransform_jupynb_v")
        batch_op.drop_index("ix_core.dtransform_name")
        batch_op.drop_index("ix_core.dtransform_pipeline_id")
        batch_op.drop_index("ix_core.dtransform_pipeline_v")
        batch_op.create_index(batch_op.f("ix_core.run_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_jupynb_id"), ["jupynb_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_jupynb_v"), ["jupynb_v"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_pipeline_id"), ["pipeline_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_pipeline_v"), ["pipeline_v"], unique=False)
        batch_op.create_foreign_key(
            batch_op.f("fk_core.run_pipeline_id_pipeline"),
            "core.pipeline",
            ["pipeline_id", "pipeline_v"],
            ["id", "v"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_core.run_jupynb_id_jupynb"),
            "core.jupynb",
            ["jupynb_id", "jupynb_v"],
            ["id", "v"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_core.run_created_by_user"),
            "core.user",
            ["created_by"],
            ["id"],
        )


def downgrade() -> None:
    pass



from pathlib import Path


def test_create_to_load():
    import lamindb.setup as lnsetup

    storage_root = "mydata-test-db"
    lnsetup.login(
        "raspbear@gmx.de",
        password="MmR4YuQEyb0yxu7dAwJZTjLzR1Az2lN4Q4IduDlO",
    )
    lnsetup.init(storage=storage_root)

    # if importing this at the top of the file lamin will try to
    # create unnecessary tables
    import lamindb as ln

    transform = ln.Transform(id="83jf", v="1", name="test", type="pipeline")
    ln.add(transform)
    run = ln.schema.Run(transform=transform)
    ln.add(run)
    ln.select(ln.schema.Storage, root=str(lnsetup.settings.instance.storage.root)).one()

    ln.schema._core.get_db_metadata_as_dict()
    table_object = ln.schema._core.get_table_object("core.file")
    ln.schema._core.get_table_metadata_as_dict(table_object)

    (Path(storage_root) / "mydata-test-db.lndb").unlink()
    # Note that this merely removes database file but doesn't clean out the instance_settings file!  # noqa
    # Hence, we need to also clean that out:
    from lndb.dev._settings_store import current_instance_settings_file

    current_instance_settings_file().unlink()


if __name__ == "__main__":
    test_create_to_load()



"""v0.28.2."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "24e55331f27c"
down_revision = "6952574e2d49"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    op.alter_column(
        f"{prefix}dobject",
        column_name="_filekey",
        new_column_name="_objectkey",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}dfolder",
        column_name="_folderkey",
        new_column_name="_objectkey",
        schema=schema,
    )

    if sqlite:
        drop_index_filekey = "ix_core.dobject__filekey"
        drop_index_folderkey = "ix_core.dfolder__folderkey"
        create_index_objectkey_dobject = "ix_core.dobject__objectkey"
        create_index_objectkey_dfolder = "ix_core.dfolder__objectkey"
    else:
        drop_index_filekey = "ix_core_dobject__filekey"
        drop_index_folderkey = "ix_core_dfolder__folderkey"
        create_index_objectkey_dobject = "ix_core_dobject__objectkey"
        create_index_objectkey_dfolder = "ix_core_dfolder__objectkey"

        with op.batch_alter_table(f"{prefix}dobject", schema=schema) as batch_op:
            batch_op.drop_index(drop_index_filekey)
            batch_op.create_index(batch_op.f(create_index_objectkey_dobject), ["_objectkey"], unique=False)
        with op.batch_alter_table(f"{prefix}dfolder", schema=schema) as batch_op:
            batch_op.drop_index(drop_index_folderkey)
            batch_op.create_index(batch_op.f(create_index_objectkey_dfolder), ["_objectkey"], unique=False)


def downgrade() -> None:
    pass



"""v0.28.3."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "3dd9f8d41861"
down_revision = "24e55331f27c"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    op.rename_table(f"{prefix}dobjects_features", f"{prefix}dobject_features", schema=schema)


def downgrade() -> None:
    pass



"""LaminDB: Manage R&D data & analyses.

Import the package::

   import lamindb as ln

The central class of the API is `File`, a wrapper for files, on-disk (`zarr`, etc.)
and in-memory data objects (`DataFrame`, `AnnData`, etc.).

.. autosummary::
   :toctree: .

   File
   Folder

Track runs of data transformations:

.. autosummary::
   :toctree: .

   Run
   Transform

Track data by feature sets:

.. autosummary::
   :toctree: .

   Features

Query & manipulate data:

.. autosummary::
   :toctree: .

   select
   add
   parse
   delete

Manipulate data with open session:

.. autosummary::
   :toctree: .

   Session

Utility functions:

.. autosummary::
   :toctree: .

   track
   view

Basic entities:

.. autosummary::
   :toctree: .

   User
   Project
   Storage

Schema - entities and their relations:

.. autosummary::
   :toctree: .

   schema

Setup:

.. autosummary::
   :toctree: .

   setup

Developer API:

.. autosummary::
   :toctree: .

   link
   context
   settings
   types
   dev
"""

__version__ = "0.39.7"  # denote a release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_logger import py_version_warning as _py_version_warning
from lndb._check_instance_setup import check_instance_setup as _check_instance_setup

_py_version_warning("3.8", "3.10")

_INSTANCE_SETUP = _check_instance_setup(from_lamindb=True)

# allow the user to call setup
from . import setup  # noqa

# only import all other functionality if setup was successful
if _INSTANCE_SETUP:
    from lnschema_core import (  # noqa
        Features,
        File,
        Folder,
        Project,
        Run,
        Storage,
        Transform,
        User,
        link,
        types,
    )

    from . import _check_versions  # executes checks during import
    from . import dev  # noqa
    from . import schema  # noqa
    from ._context import context  # noqa

    track = context._track  # noqa
    from lndb_storage import subset

    # deprecated
    from lndb_storage.object import lazy

    from . import _amend_file  # noqa
    from . import _amend_folder  # noqa
    from ._delete import delete  # noqa
    from ._nb import nb  # noqa
    from ._parse import parse  # noqa
    from ._settings import settings
    from ._view import view  # noqa
    from .dev.db import Session  # noqa
    from .dev.db._add import add  # noqa
    from .dev.db._select import select  # noqa
